import "Turbine";
import "Turbine.Gameplay";
import "Turbine.UI";
import "Turbine.UI.Lotro";

ColorBar = class(Turbine.UI.Control);

function ColorBar:Constructor(n, w, h)
	Turbine.UI.Control.Constructor(self);

	self:SetBackColor(Turbine.UI.Color(1, 0.0, 0.0, 0.0));
	self:SetSize(w, h);

	self.n = n;
	self.w = w;
	self.backBars = { };
	self.frontBars = { };
	for i = 1, n, 1 do
		self.backBars[i] = Turbine.UI.Control();
		self.backBars[i]:SetParent(self);
		self.backBars[i]:SetPosition(0, 0);
		self.backBars[i]:SetSize(w, h);
		self.backBars[i]:SetZOrder(i);
		self.frontBars[i] = Turbine.UI.Control();
		self.frontBars[i]:SetParent(self);
		self.frontBars[i]:SetPosition(0, 0);
		self.frontBars[i]:SetSize(w, h);
		self.frontBars[i]:SetZOrder(n + i);
	end
end

function ColorBar:SetBarColor(i, c)
	if i > self.n then
		return;
	end

	self.backBars[i]:SetBackColor(Turbine.UI.Color(0.5, (c.R + 0.5) / 2.0, (c.G + 0.5) / 2.0, (c.B + 0.5) / 2.0));
	self.frontBars[i]:SetBackColor(c);
end

function ColorBar:SetStats(i, l, c, r, m)
	if i > self.n then
		return;
	end

	if r <= l then
		self.backBars[i]:SetVisible(false);
		self.frontBars[i]:SetVisible(false);
		return;
	end

	m = math.max(m, 1);
	l = math.min(l, m);
	r = math.min(r, m);
	c = math.max(l, c);
	c = math.min(r, c);
	l = math.floor((l * self.w / m) + 0.5)
	c = math.floor((c * self.w / m) + 0.5)
	r = math.floor((r * self.w / m) + 0.5)

	if (r - l) <= 0 then
		self.backBars[i]:SetVisible(false);
		self.frontBars[i]:SetVisible(false);
	else
		self.backBars[i]:SetLeft(l);
		self.backBars[i]:SetWidth(r - l);
		self.backBars[i]:SetVisible(true);
		if (c - l) <= 0 then
			self.frontBars[i]:SetVisible(false);
		else
			self.frontBars[i]:SetLeft(l);
			self.frontBars[i]:SetWidth(c - l);
			self.frontBars[i]:SetVisible(true);
		end
	end
end

function ColorBar:SetMouseVisible(b)
	Turbine.UI.Control.SetMouseVisible(self, b);
	for i = 1, self.n, 1 do
		self.backBars[i]:SetMouseVisible(b);
		self.frontBars[i]:SetMouseVisible(b);
	end
end