import "Turbine";
import "Turbine.Gameplay";
import "Turbine.UI";
import "Turbine.UI.Lotro";
import "moebius92.ColorBar";
import "moebius92.Handlers";
import "moebius92.Misc";

-- Note: Does not handle self referencing tables
function tabledeepcopy(t)
	local r = { };

	for k, v in pairs(t) do
		if type(v) == "table" then
			r[k] = tabledeepcopy(v);
		else
			r[k] = v;
		end
	end

	return r;
end

tempMoraleEffects = { };
tempMoraleEffects["Word of Exaltation"] = true;
tempMoraleEffects["Joy of Battle"] = true;
tempMoraleEffects["Glory of Battle"] = true;
tempMoraleEffects["Sudden Defence"] = true;
tempMoraleEffects["Song of the Hammerhand"] = true;
tempMoraleEffects["Gift of the Hammerhand"] = true;

local baseDurationImageID = 0x41007E35;

effectGroup = { };
effectSlot = { };

categoryString = { };
categoryString[Turbine.Gameplay.EffectCategory.Fear] = "fear";
categoryString[Turbine.Gameplay.EffectCategory.Wound] = "wound";
categoryString[Turbine.Gameplay.EffectCategory.Disease] = "disease";
categoryString[Turbine.Gameplay.EffectCategory.Poison] = "poison";

categoryColor = { };
categoryColor[Turbine.Gameplay.EffectCategory.Fear] = Turbine.UI.Color(1.0, 1.0, 0.0, 1.0);
categoryColor[Turbine.Gameplay.EffectCategory.Wound] = Turbine.UI.Color(1.0, 1.0, 0.0, 0.0);
categoryColor[Turbine.Gameplay.EffectCategory.Disease] = Turbine.UI.Color(1.0, 1.0, 1.0, 0.0);
categoryColor[Turbine.Gameplay.EffectCategory.Poison] = Turbine.UI.Color(1.0, 0.0, 1.0, 0.0);

function EffectSlot(list, debuff, icon, name, category, curable)
	local r;

	debuff = debuff and "D" or "B";

	r = list[debuff .. "/" .. icon .. "/" .. name];
	if r ~= nil then
		return r;
	end

	r = list[debuff .. "/" .. name];
	if r ~= nil then
		return r;
	end

	r = list[debuff .. "/" .. icon];
	if r ~= nil then
		return r;
	end

	if categoryString[category] ~= nil then
		curable = curable and "curable" or "incurable";

		r = list[debuff .. "/<" .. categoryString[category] .. "-" .. curable .. ">"];
		if r ~= nil then
			return r;
		end

		r = list[debuff .. "/<" .. categoryString[category] .. ">"];
		if r ~= nil then
			return r;
		end

		r = list[debuff .. "/<" .. curable .. ">"];
		if r ~= nil then
			return r;
		end
	end

	return nil;
end

classColor = { };
classColor[Turbine.Gameplay.Class.Burglar] =		Turbine.UI.Color(0.3, 0.0, 0.0, 0.0);
classColor[Turbine.Gameplay.Class.Captain] =		Turbine.UI.Color(0.3, 0.0, 0.0, 1.0);
classColor[Turbine.Gameplay.Class.Champion] =		Turbine.UI.Color(0.3, 1.0, 0.0, 0.0);
classColor[Turbine.Gameplay.Class.Guardian] =		Turbine.UI.Color(0.3, 1.0, 0.5, 0.0);
classColor[Turbine.Gameplay.Class.Hunter] =			Turbine.UI.Color(0.3, 0.0, 1.0, 0.0);
classColor[Turbine.Gameplay.Class.LoreMaster] =		Turbine.UI.Color(0.3, 0.0, 1.0, 1.0);
classColor[Turbine.Gameplay.Class.Minstrel] =		Turbine.UI.Color(0.3, 1.0, 1.0, 0.0);
classColor[Turbine.Gameplay.Class.RuneKeeper] =		Turbine.UI.Color(0.3, 1.0, 1.0, 1.0);
classColor[Turbine.Gameplay.Class.Warden] =			Turbine.UI.Color(0.3, 1.0, 0.0, 1.0);

readyColor = { };
readyColor[Turbine.Gameplay.ReadyState.Ready] = Turbine.UI.Color(1.0, 0.0, 1.0, 0.0);
readyColor[Turbine.Gameplay.ReadyState.NotReady] = Turbine.UI.Color(1.0, 1.0, 0.0, 0.0);

PartyMemberWindow = class(Turbine.UI.Window);

function PartyMemberWindow:Constructor(parent, actor, size, width, t)
	Turbine.UI.Window.Constructor(self);

	self.parent = parent;
	self.actor = actor;
	self.t = t;
	if self.t == nil then
		self.class = actor:GetClass();
	else
		self.class = t;
	end
	self.size = size;
	self.width = width;
	self.dead = false;
	self.damage = 0;
	self.rolls = { };
	self.rollOrder = nil;
	self.actorHandle = { };

	self.select = Turbine.UI.Lotro.EntityControl();
	self.select:SetParent(self);
	self.select:SetPosition(0, 0);
	self.select:SetMouseVisible(true);
	self.select:SetEntity(actor);
	self.select:SetSelectionEnabled(true);
	self.select:SetContextMenuEnabled(true);

	if classColor[self.class] ~= nil then
		self.color = classColor[self.class];
	else
		self.color = Turbine.UI.Color(0.3, 1.0, 1.0, 1.0)
	end
	self:SetBackColor(self.color);

	self.name = Turbine.UI.Label();
	self.name:SetOutlineColor(Turbine.UI.Color(1.0, 0.0, 0.0, 0.0));
	self.name:SetFontStyle(Turbine.UI.FontStyle.Outline);
	self.name:SetFont(Turbine.UI.Lotro.Font.Verdana16);
	self.name:SetMultiline(false);
	self.name:SetZOrder(100);
	self.name:SetParent(self);
	self.name:SetMouseVisible(false);
	self.actorHandle["IsLinkDeadChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "IsLinkDeadChanged", function(sender, args) self:UpdateName(); end);
	self.actorHandle["IsVoiceEnabledChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "IsVoiceEnabledChanged", function(sender, args) self:UpdateName(); end);
	self.actorHandle["NameChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "NameChanged", function(sender, args) self:UpdateName(); end);
	self:UpdateName();

	if self.t == nil then
		self.ready = Turbine.UI.Control();
		self.ready:SetZOrder(0);
		self.ready:SetParent(self);
		self.ready:SetVisible(true);
		self.ready:SetMouseVisible(false);
		self.actorHandle["ReadyStateChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "ReadyStateChanged", function(sender, args) self:UpdateReadyState(); end);
		self.actorHandle["IsVoiceActiveChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "IsVoiceActiveChanged", function(sender, args) self:UpdateReadyState(); end);
		self:UpdateReadyState();

		self.actorHandle["PetChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "PetChanged", function(sender, args) self.parent:PartyMemberChanged(); end);
	else
		self.triangle = Turbine.UI.Control();
		self.triangle:SetZOrder(100);
		self.triangle:SetParent(self);
		self.triangle:SetPosition(0, 0);
		self.triangle:SetSize(size, size);
		self.triangle:SetVisible(true);
		self.triangle:SetMouseVisible(false);
		self.triangle:SetBackground("moebius92/PartyFrame/triangle.tga");
		self.triangle:SetBackColor(Turbine.UI.Color(1.0, self.color.R, self.color.G, self.color.B));
		self.triangle:SetBackColorBlendMode(Turbine.UI.BlendMode.Multiply);
	end

	self.threat = 0;
	self.oldMorale = actor:GetMorale();
	self.moraleBar = moebius92.ColorBar.ColorBar(6, width - 2, (size - 6) / 2);
	self.moraleBar.tempEnd = 0;
	self.moraleBar.tempEffect = { };
	self.moraleBar:SetBarColor(1, Turbine.UI.Color(1, 0.0, 1.0, 0.0));
	self.moraleBar:SetBarColor(2, Turbine.UI.Color(1, 0.5, 0.0, 1.0));
	self.moraleBar:SetBarColor(3, Turbine.UI.Color(1, 1.0, 0.5, 0.0));
	self.moraleBar:SetBarColor(4, Turbine.UI.Color(1, 0.5, 1.0, 0.5));
	self.moraleBar:SetBarColor(5, Turbine.UI.Color(1, 1.0, 0.0, 0.0));
	self.moraleBar:SetParent(self);
	self.moraleBar:SetMouseVisible(false);
	self.actorHandle["MoraleChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "MoraleChanged", function(sender, args) self:UpdateMorale(); end);
	self.actorHandle["MaxMoraleChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "MaxMoraleChanged", function(sender, args) self:UpdateMorale(); end);
	self.actorHandle["BaseMaxMoraleChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "BaseMaxMoraleChanged", function(sender, args) self:UpdateMorale(); end);
	self.actorHandle["TemporaryMoraleChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "TemporaryMoraleChanged", function(sender, args) self:UpdateMorale(); end);
	self.actorHandle["MaxTemporaryMoraleChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "MaxTemporaryMoraleChanged", function(sender, args) self:UpdateMorale(); end);
	self:UpdateMorale();

	self.powerBar = moebius92.ColorBar.ColorBar(4, width - 2, (size - 6) / 2);
	self.powerBar.tempEnd = 0;
	self.powerBar.tempEffect = { };
	self.powerBar:SetBarColor(1, Turbine.UI.Color(1, 0.0, 0.0, 1.0));
	self.powerBar:SetBarColor(2, Turbine.UI.Color(1, 0.5, 1.0, 0.0));
	self.powerBar:SetBarColor(3, Turbine.UI.Color(1, 1.0, 0.0, 0.5));
	self.powerBar:SetBarColor(4, Turbine.UI.Color(1, 0.5, 0.5, 1.0));
	self.powerBar:SetParent(self);
	self.powerBar:SetMouseVisible(false);
	self.actorHandle["PowerChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "PowerChanged", function(sender, args) self:UpdatePower(); end);
	self.actorHandle["MaxPowerChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "MaxPowerChanged", function(sender, args) self:UpdatePower(); end);
	self.actorHandle["BaseMaxPowerChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "BaseMaxPowerChanged", function(sender, args) self:UpdatePower(); end);
	self.actorHandle["TemporaryPowerChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "TemporaryPowerChanged", function(sender, args) self:UpdatePower(); end);
	self.actorHandle["MaxTemporaryPowerChanged"] = moebius92.Handlers.AddEventHandler(self.actor, "MaxTemporaryPowerChanged", function(sender, args) self:UpdatePower(); end);
	self:UpdatePower();

	self.effects = self.actor:GetEffects();
	self.effectsHandle = { };

	self.effectSlots = { };
	for i = -5, 5, 1 do
		if i ~= 0 then
			self.effectSlots[i] = { };
			self.effectSlots[i].icon = Turbine.UI.Control();
			self.effectSlots[i].icon:SetZOrder(101);
			self.effectSlots[i].icon:SetParent(self);
			self.effectSlots[i].icon:SetSize(size, size);
			self.effectSlots[i].icon:SetVisible(false);
			self.effectSlots[i].icon:SetBackColor(Turbine.UI.Color(0.7, 1.0, 0.0, 0.0));
			self.effectSlots[i].icon:SetBackColorBlendMode(Turbine.UI.BlendMode.Multiply);
			self.effectSlots[i].display = Turbine.UI.Lotro.EffectDisplay();
			self.effectSlots[i].display:SetZOrder(101);
			self.effectSlots[i].display:SetParent(self);
			self.effectSlots[i].display:SetSize(size, size);
			self.effectSlots[i].display:SetVisible(false);
			self.effectSlots[i].duration = Turbine.UI.Control();
			self.effectSlots[i].duration:SetPosition(0, 0);
			self.effectSlots[i].duration:SetSize(size, size);
			self.effectSlots[i].duration:SetZOrder(150);
			self.effectSlots[i].duration:SetOpacity(0.7);
			self.effectSlots[i].duration:SetBackColor(Turbine.UI.Color(1.0, 0.0, 0.0, 0.0));
			self.effectSlots[i].duration:SetBackColorBlendMode(Turbine.UI.BlendMode.Multiply);
			self.effectSlots[i].duration:SetMouseVisible(false);
			self.effectSlots[i].duration:SetParent(self.effectSlots[i].display);
			self.effectSlots[i].text = Turbine.UI.Label();
			self.effectSlots[i].text:SetOutlineColor(Turbine.UI.Color(1.0, 0.0, 0.0, 0.0));
			self.effectSlots[i].text:SetFontStyle(Turbine.UI.FontStyle.Outline);
			self.effectSlots[i].text:SetFont(Turbine.UI.Lotro.Font.Verdana12);
			self.effectSlots[i].text:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
			self.effectSlots[i].text:SetPosition(0, 0);
			self.effectSlots[i].text:SetSize(size, size);
			self.effectSlots[i].text:SetZOrder(200);
			self.effectSlots[i].text:SetStretchMode(1);
			self.effectSlots[i].text:SetMouseVisible(false);
			self.effectSlots[i].text:SetParent(self.effectSlots[i].display);
			self.effectSlots[i].effects = { };
		end
	end

	self.effectsHandle["EffectAdded"] = moebius92.Handlers.AddEventHandler(self.effects, "EffectAdded", function(sender, args) self:AddEffect(self.effects:Get(args.Index)); end);
	self.effectsHandle["EffectRemoved"] = moebius92.Handlers.AddEventHandler(self.effects, "EffectRemoved", function(sender, args) self:RemoveEffect(args.Effect, true); end);
	self.effectsHandle["EffectsCleared"] = moebius92.Handlers.AddEventHandler(self.effects, "EffectsCleared", function(sender, args) self:ClearEffects(); end);

	self:Layout(self.parent.settings.layout);

	self:SetVisible(self.parent.hide);
end

function PartyMemberWindow:Layout(layout)
	if layout == 1 then
		self:SetSize(self.width + self.size * 11, self.size);

		self.name:SetPosition(self.size, 0);
		self.name:SetSize(self.size * 5, self.size);
		self.name:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);

		if self.t == nil then
			self.ready:SetPosition(self.size / 4, self.size / 4);
			self.ready:SetSize(self.size / 2, self.size / 2);
		else
			self.triangle:SetVisible(true);
		end

		self.moraleBar:SetPosition(self.size * 6 + 1, 3);
		self.powerBar:SetPosition(self.size * 6 + 1, self.size / 2);

		for i = -5, 5, 1 do
			if i < 0 then
				self.effectSlots[i].icon:SetPosition((i + 6) * self.size, 0);
				self.effectSlots[i].display:SetPosition((i + 6) * self.size, 0);
			elseif i > 0 then
				self.effectSlots[i].icon:SetPosition(self.width + (i + 5) * self.size, 0);
				self.effectSlots[i].display:SetPosition(self.width + (i + 5) * self.size, 0);
			end
		end
	else
		self:SetSize(self.size * 5, self.size * 3);

		self.name:SetPosition(2, 0);
		self.name:SetSize(self.size * 5 - 4, self.size);
		self.name:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);

		if self.t == nil then
			self.ready:SetPosition(0, self.size);
			self.ready:SetSize(self.size * 5, self.size);
		else
			self.triangle:SetVisible(false);
		end

		self.moraleBar:SetPosition((self.size * 5 - self.width) / 2, self.size + 3);
		self.powerBar:SetPosition((self.size * 5 - self.width) / 2, self.size * 1.5);

		for i = -5, 5, 1 do
			if i < 0 then
				self.effectSlots[i].icon:SetPosition((i + 5) * self.size, 0);
				self.effectSlots[i].display:SetPosition((i + 5) * self.size, 0);
			elseif i > 0 then
				self.effectSlots[i].icon:SetPosition((i - 1) * self.size, self.size * 2);
				self.effectSlots[i].display:SetPosition((i - 1) * self.size, self.size * 2);
			end
		end
	end

	self.select:SetSize(self:GetSize());
	self:SetupEffects();
end

function PartyMemberWindow:SetupEffects()
	self:ClearEffects();
	self.lastEffectUpdate = 0;
	for i = 1, self.effects:GetCount(), 1 do
		self:AddEffect(self.effects:Get(i));
	end
end

function PartyMemberWindow:MouseDown(args)
	if args.Button == Turbine.UI.MouseButton.Left then
		self.dragging = true;
		self.safeY = nil;
	end
end

function IsSafeY(self, Y)
	if self.safeY ~= nil then
		if self.safeY > 0 then
			if Y > 0 and Y < self.safeY then
				self.safeY = Y;
				return false;
			end
		elseif self.safeY < 0 then
			if Y < 0 and Y > self.safeY then
				self.safeY = Y;
				return false;
			end
		end
		self.safeY = nil;
	end
	return true;
end

function PartyMemberWindow:MouseMove(args)
	local w, h = self:GetSize();

	if self.dragging then
		local x, y;

		if not IsSafeY(self, args.Y) then
			return;
		end

		x = math.floor(args.X / w);
		y = math.floor(args.Y / h);
		if x == 0 and y == 1 and self.parent.settings.layout == 1 and self.pet ~= nil then
			y = 0;
		end

		if x ~= 0 or y ~= 0 then
			if self.parent:MoveMember(self.id, x, y) and self.parent.settings.layout == 1 then
				if y < 0 then
					self.safeY = math.huge;
				else
					self.safeY = -math.huge;
				end
			end
		end
	end
end

function PartyMemberWindow:MouseUp(args)
	if args.Button == Turbine.UI.MouseButton.Left then
		self.dragging = false;
	end
end

function PartyMemberWindow:MouseClick(args)
	if args.Button == Turbine.UI.MouseButton.Right then
		self.parent.memberSelect = self;
		for i = 1, #self.parent.memberOptions, 1 do
			self.parent.memberOptions[i]:SetChecked(i == self.showPet);
			if self.parent.settings.layout == 1 then
				self.parent.memberOptions[i]:SetEnabled(i <= 2);
			else
				self.parent.memberOptions[i]:SetEnabled(true);
			end
		end
		self.parent.memberMenu:ShowMenu();
	end
end

function PartyMemberWindow:RemoveHandlers()
	for k, v in pairs(self.actorHandle) do
		moebius92.Handlers.RemoveEventHandler(self.actor, k, v);
	end
	for k, v in pairs(self.effectsHandle) do
		moebius92.Handlers.RemoveEventHandler(self.effects, k, v);
	end
end

function PartyMemberWindow:UpdateName()
	local x = 1.0;
	local name = self.actor:GetName();

	if name == nil or name == "" then
		self.pollName = true;
	else
		self.pollName = false;
	end

	if self.dead then
		x = 0.5;
	end

	if self.rollOrder == nil then
		self.name:SetText(name);
	else
		self.name:SetText("(" .. self.rollOrder .. ") " .. name);
	end

	if self.t ~= nil then
		self.name:SetForeColor(Turbine.UI.Color(1.0, x, x, x));
	elseif self.actor:IsLinkDead() then
		self.name:SetForeColor(Turbine.UI.Color(1.0, x, 0.0, 0.0));
	elseif self.actor:IsVoiceEnabled() then
		self.name:SetForeColor(Turbine.UI.Color(1.0, 0.0, x, 0.0));
	else
		self.name:SetForeColor(Turbine.UI.Color(1.0, x, x, x));
	end
end

function PartyMemberWindow:UpdateReadyState()
	local r = self.actor:GetReadyState();
	local v = self.actor:IsVoiceActive();
	local c = readyColor[r];

	if c == nil then
		if v then
			c = Turbine.UI.Color(1.0, 0.0, 0.0, 1.0);
		else
			c = Turbine.UI.Color(1.0, 1.0, 1.0, 1.0);
		end
	elseif v then
		c = Turbine.UI.Color(1.0, c.R / 2, c.G / 2, 1.0);
	end

	if self.leader then
		self.ready:SetBackColor(c);
	else
		self.ready:SetBackColor(Turbine.UI.Color(1.0, c.R / 2.0, c.G / 2.0, c.B / 2.0));
	end
end

function PartyMemberWindow:UpdateMorale()
	local cur = self.actor:GetMorale();
	local base = self.actor:GetBaseMaxMorale();
	local max = self.actor:GetMaxMorale();
	local temp = self.actor:GetTemporaryMorale();
	local tmax = self.actor:GetMaxTemporaryMorale();
	local damage = self.oldMorale - cur;

	if cur == 0 then
		if not self.dead then
			self.dead = true;
			self:UpdateName();
		end
	else
		if self.dead then
			self.dead = false;
			self:UpdateName();
		end
	end

	if self.parent.inCombat and damage > 0 then
		local time = Turbine.Engine.GetGameTime();

		if self.parent.threatTime > 0 then
			local item = { };

			self.damage = self.damage + damage;
			item.damage = damage;
			item.time = time;
			if self.damageTail ~= nil then
				self.damageTail.next = item;
				self.damageTail = item;
			else
				self.damageHead = item;
				self.damageTail = item;
			end

			while self.damageHead ~= nil do
				if (self.damageHead.time + self.parent.threatTime) >= time then
					break;
				end
				self.damage = self.damage - self.damageHead.damage;
				self.damageHead = self.damageHead.next;
			end

			if self.threat < self.damage then
				self.threat = self.damage;
			end
		end
	end

	local m = math.max(max, base) + temp;
	local min = math.min(max, base);

	self.maxMorale = math.max(max, base) + math.min(tmax, base);
	self.oldMorale = cur;
	
	local t = 1.0;
	if self.moraleBar.tempEnd ~= 0 then
		t = math.min(10.0, math.max(0, self.moraleBar.tempEnd - Turbine.Engine.GetGameTime())) / 10.0;
	end
	self.moraleBar:SetBarColor(2, Turbine.UI.Color(1, 0.5, 0.0, t));

	local mp = math.max(m, self.parent.maxMorale);
	self.moraleBar:SetStats(1, 0, cur, min, mp);
	self.moraleBar:SetStats(2, max, max + temp, max + temp, mp);
	self.moraleBar:SetStats(3, max + temp, m, m, mp);
	self.moraleBar:SetStats(4, base, cur, max, mp);
	self.moraleBar:SetStats(5, 0, cur, math.min(self.threat, max), mp);
	self.parent:UpdateMorale();
end

function PartyMemberWindow:UpdatePower()
	local cur = self.actor:GetPower();
	local base = self.actor:GetBaseMaxPower();
	local max = self.actor:GetMaxPower();
	local temp = self.actor:GetTemporaryPower();
	local tmax = self.actor:GetMaxTemporaryPower();

	local m = math.max(max, base) + temp;

	self.maxPower = math.max(max, base) + math.min(tmax, base);

	local t = 1.0;
	if self.powerBar.tempEnd ~= 0 then
		t = math.min(10.0, math.max(0, self.powerBar.tempEnd - Turbine.Engine.GetGameTime())) / 10.0;
	end
	self.powerBar:SetBarColor(2, Turbine.UI.Color(1, 0.5, t, 0.0));

	local mp = math.max(m, self.parent.maxPower);
	self.powerBar:SetStats(1, 0, cur, max, mp);
	self.powerBar:SetStats(2, max, max + temp, max + temp, mp);
	self.powerBar:SetStats(3, max + temp, m, m, mp);
	self.powerBar:SetStats(4, math.max(0, max - self.parent.powerFeed), cur, max, mp);
	self.parent:UpdatePower();
end

function PartyMemberWindow:UpdateEffectTime(i, time)
	if self.effectSlots[i].durationTime == nil or self.effectSlots[i].durationTime > 1000000 then
		self.effectSlots[i].duration:SetVisible(false);
		return false;
	else
		local x = self.parent.options.slotsButton[i].duration;

		self.effectSlots[i].duration:SetVisible(true);
		if x > 0 then
			x = math.floor(math.min(59, math.max(0, 60.0 * (time - (self.effectSlots[i].startTime + self.effectSlots[i].durationTime - x)) / x)));
		else
			x = math.floor(math.min(59, math.max(0, 60.0 * (time - self.effectSlots[i].startTime) / self.effectSlots[i].durationTime)));
		end
		if x ~= self.effectSlots[i].durationIcon then
			self.effectSlots[i].duration:SetBackground(baseDurationImageID + x);
			self.effectSlots[i].duration:SetStretchMode(2);
			self.effectSlots[i].duration:SetStretchMode(1);
			self.effectSlots[i].duration:SetSize(self.size, self.size);
			self.effectSlots[i].durationIcon = x;
		end
		return true;
	end
end

function PartyMemberWindow:UpdateEffect(i)
	local item = self.effectSlots[i].effects[1];

	self.effectSlots[i].durationTime = nil;

	if effectGroup[i] ~= nil then
		local found;

		for _, v in ipairs(effectGroup[i]) do
			found = false;
			for _, v1 in pairs(self.effectSlots[i].effects) do
				if EffectSlot(v.effects, v1.effect:IsDebuff(), v1.effect:GetIcon(), v1.effect:GetName(), v1.effect:GetCategory(), v1.effect:IsCurable()) then
					found = true;
					break;
				end
			end

			if not found then
				self.effectSlots[i].icon:SetVisible(true);
				self.effectSlots[i].icon:SetBackground(v.icon);
				self.effectSlots[i].icon:SetStretchMode(2);
				self.effectSlots[i].icon:SetStretchMode(1);
				self.effectSlots[i].icon:SetSize(self.size, self.size);
				break;
			end
		end

		if found then
			self.effectSlots[i].icon:SetVisible(false);
		end

		self.effectSlots[i].display:SetVisible(false);
	elseif item ~= nil then
		local cat = item.effect:GetCategory();
		local c = categoryColor[cat];
		local n = #self.effectSlots[i].effects;
		local name = item.effect:GetName();
		local t;

		self.effectSlots[i].display:SetEffect(item.effect);
		self.effectSlots[i].startTime = item.effect:GetStartTime();
		self.effectSlots[i].durationTime = item.effect:GetDuration();
		self.effectSlots[i].debuff = item.effect:IsDebuff();

		if not item.effect:IsDebuff() or not item.effect:IsCurable() then
			c = nil;
		end

		if c ~= nil then
			n = 0;
			for _, v in pairs(self.effectSlots[i].effects) do
				if v.effect:IsDebuff() and v.effect:IsCurable() and v.effect:GetCategory() == cat then
					n = n + 1;
				end
			end
			if n > 1 then
				self.effectSlots[i].text:SetText(tostring(n));
				self.effectSlots[i].text:SetForeColor(c);
				self.effectSlots[i].text:SetVisible(true);
			else
				self.effectSlots[i].text:SetVisible(false);
			end
		else
			local t = string.match(name, " ([0-9]+)$") or string.match(name, "^Tier ([0-9]+)");

			if t ~= nil then
				self.effectSlots[i].text:SetText(t);
				self.effectSlots[i].text:SetForeColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
				self.effectSlots[i].text:SetVisible(true);
			elseif n > 1 then
				self.effectSlots[i].text:SetText(tostring(n));
				self.effectSlots[i].text:SetForeColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
				self.effectSlots[i].text:SetVisible(true);
			else
				self.effectSlots[i].text:SetVisible(false);
			end
		end
		self:UpdateEffectTime(i, Turbine.Engine.GetGameTime());
		self.effectSlots[i].display:SetVisible(true);
		self.effectSlots[i].icon:SetVisible(false);
	else
		self.effectSlots[i].display:SetEffect(nil);
		self.effectSlots[i].display:SetVisible(false);
		self.effectSlots[i].icon:SetVisible(false);
	end
end

function PartyMemberWindow:AddEffect(effect)
	local id;
	local time = Turbine.Engine.GetGameTime();
	local name = effect:GetName();

	self.parent:LogEffect(effect);
	self:SetWantsUpdates(true);

	self.lastEffectUpdate = time;
	for i = 1, self.effects:GetCount(), 1 do
		id = self.effects:Get(i):GetID();

		if self.effectID[id] ~= nil then
			self.effectID[id].time = time;
		end
	end

	id = effect:GetID();

	if tempMoraleEffects[name] then
		self.moraleBar.tempEffect[id] = effect:GetStartTime() + effect:GetDuration();
		self.moraleBar.tempEnd = 0;
		for _, v in pairs(self.moraleBar.tempEffect) do
			self.moraleBar.tempEnd = math.max(self.moraleBar.tempEnd, v);
		end
	end

	if self.effectID[id] == nil then
		local slot = EffectSlot(effectSlot, effect:IsDebuff(), effect:GetIcon(), name, effect:GetCategory(), effect:IsCurable());

		if slot ~= nil then
			local item = { };

			item.effect = effect;
			item.time = time;
			item.slot = slot;
			item.startTime = effect:GetStartTime();
			item.durationTime = effect:GetDuration();
			item.endTime = item.startTime + item.durationTime;
			table.insert(self.effectSlots[slot].effects, item);
			table.sort(self.effectSlots[slot].effects, function(a, b) return a.startTime > b.startTime; end);
			self.effectID[id] = item;
			self:UpdateEffect(slot);
		end
	end
end

function PartyMemberWindow:RemoveEffect(effect, mark)
	local id;
	local time = Turbine.Engine.GetGameTime();
	local name = effect:GetName();

	if mark then
		self.lastEffectUpdate = time;
		for i = 1, self.effects:GetCount(), 1 do
			id = self.effects:Get(i):GetID();

			if self.effectID[id] ~= nil then
				self.effectID[id].time = time;
			end
		end
	end

	id = effect:GetID();

	if self.moraleBar.tempEffect[id] ~= nil then
		self.moraleBar.tempEffect[id] = nil;
		self.moraleBar.tempEnd = 0;
		for _, v in pairs(self.moraleBar.tempEffect) do
			self.moraleBar.tempEnd = math.max(self.moraleBar.tempEnd, v);
		end
	end

	if self.effectID[id] ~= nil then
		local slot = self.effectID[id].slot;

		for k, v in pairs(self.effectSlots[slot].effects) do
			if v == self.effectID[id] then
				table.remove(self.effectSlots[slot].effects, k);
				break;
			end
		end
		self.effectID[id] = nil;
		self:UpdateEffect(slot);
	end
end

function PartyMemberWindow:ClearEffects()
	for i = -5, 5, 1 do
		if i ~= 0 then
			self.effectSlots[i].effects = { };
			self:UpdateEffect(i);
		
		end
	end
	self.effectID = { };

	self.moraleBar.tempEffect = { };
	self.moraleBar.tempEnd = 0;
	self.powerBar.tempEffect = { };
	self.powerBar.tempEnd = 0;
end

function PartyMemberWindow:Update()
	local time = Turbine.Engine.GetGameTime();
	local update = false;

	if self.moraleBar.tempEnd ~= 0 then
		self:UpdateMorale();
		update = true;
	end

	if self.powerBar.tempEnd ~= 0 then
		self:UpdatePower();
		update = true;
	end

	for i = -5, 5, 1 do
		if i ~= 0 and #self.effectSlots[i].effects > 0 then
			if self:UpdateEffectTime(i, time) then
				update = true;
			end
		end
	end

	if self.lastEffectUpdate == time then
		for k, v in pairs(self.effectID) do
			if v.time ~= time then
				self:RemoveEffect(v.effect, false);
			end
		end
	else
		for k, v in pairs(self.effectID) do
			if v.endTime <= time then
				self:RemoveEffect(v.effect, false);
			end
		end
	end

	for k, v in pairs(self.effectID) do
		if v.durationTime > 1000000 then
			update = true;
			break;
		end
	end

	if self.pollName then
		self:UpdateName();
		update = true;
	end

	if not update then
		self:SetWantsUpdates(false);
	end
end

PartyFrameFTab = class(Turbine.UI.Window);

function PartyFrameFTab:Constructor(parent, index)
	Turbine.UI.Window.Constructor(self);

	self.parent = parent;
	self.index = index;
	self.dragging = false;

	self:SetBackColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
	self.u = Turbine.UI.Control();
	self.u:SetParent(self);
	self.u:SetPosition(0, 0);
	self.u:SetSize(parent.size, parent.size);
	self.u:SetBackColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
	self.u:SetBackColorBlendMode(Turbine.UI.BlendMode.Multiply);
	self.u:SetMouseVisible(false);
	self.d = Turbine.UI.Control();
	self.d:SetParent(self);
	self.d:SetSize(parent.size, parent.size);
	self.d:SetBackColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
	self.d:SetBackColorBlendMode(Turbine.UI.BlendMode.Multiply);
	self.d:SetMouseVisible(false);
end

function PartyFrameFTab:Orient(right)
	if self.l > 0 then
		self:SetSize(self.parent.size, self.parent.size * self.l);
		if self.l == 1 then
			self.d:SetVisible(false);
		else
			self.d:SetPosition(0, self.parent.size * (self.l - 1));
			self.d:SetVisible(true);
		end
		if right then
			self.u:SetBackground("moebius92/PartyFrame/ur.tga");
			self.d:SetBackground("moebius92/PartyFrame/lr.tga");
		else
			self.u:SetBackground("moebius92/PartyFrame/ul.tga");
			self.d:SetBackground("moebius92/PartyFrame/ll.tga");
		end
		self:SetVisible(self.parent.move);
	else
		self:SetVisible(false);
	end
end

function PartyFrameFTab:MouseDown(args)
	self.dragging = true;
	self.safeY = nil;
end

orientDFellowship = { };
orientDFellowship[1] = -1;
orientDFellowship[2] = 1;

function PartyFrameFTab:MouseMove(args)
	if self.dragging then
		if not IsSafeY(self, args.Y) then
			return;
		end

		if args.Y < -(self.parent.size / 2) then
			local i = self.index - orientDFellowship[self.parent.settings.orient];

			if self.parent.fellowtabs[i] ~= nil then
				local x = self.parent.fellowship[self.index];

				self.parent.fellowship[self.index] = self.parent.fellowship[i];
				self.parent.fellowship[i] = x;
				self.parent.fellowtabs[self.index] = self.parent.fellowtabs[i];
				self.parent.fellowtabs[self.index].index = self.index;
				self.parent.fellowtabs[i] = self;
				self.index = i;
				self.parent:GridLayout();
				self.safeY = math.huge;
			end
		elseif args.Y > (self:GetHeight() + 0.5 * self.parent.size) then
			local i = self.index + orientDFellowship[self.parent.settings.orient];

			if self.parent.fellowtabs[i] ~= nil then
				local x = self.parent.fellowship[self.index];

				self.parent.fellowship[self.index] = self.parent.fellowship[i];
				self.parent.fellowship[i] = x;
				self.parent.fellowtabs[self.index] = self.parent.fellowtabs[i];
				self.parent.fellowtabs[self.index].index = self.index;
				self.parent.fellowtabs[i] = self;
				self.index = i;
				self.parent:GridLayout();
				self.safeY = -math.huge;
			end
		end
	end
end

function PartyFrameFTab:MouseUp(args)
	self.dragging = false;
end

petOptions = { };
petOptions[1] = "Hide Pet";
petOptions[2] = "Pet Below";
petOptions[3] = "Pet Above";
petOptions[4] = "Pet Left";
petOptions[5] = "Pet Right";

layoutOptions = { };
layoutOptions[1] = 1;
layoutOptions[2] = 1;
layoutOptions[3] = 2;
layoutOptions[4] = 2;
layoutOptions[5] = 2;
layoutOptions[6] = 2;

orientOptions = { };
orientOptions[1] = 1;
orientOptions[2] = 2;
orientOptions[3] = 1;
orientOptions[4] = 2;
orientOptions[5] = 3;
orientOptions[6] = 4;

layoutName = { };
layoutName[1] = "Line";
layoutName[2] = "Grid";

orientName = { };
orientName[1] = "Up";
orientName[2] = "Down";
orientName[3] = "Right";
orientName[4] = "Left";

PartyFrameWindow = class(Turbine.UI.Window);

function PartyFrameWindow:Constructor(size, width, version)
	Turbine.UI.Window.Constructor(self);

	self.hide = false;
	self.move = false;
	self.dragging = false;
	self.size = size;
	self.width = width;
	self.w = Turbine.UI.Display:GetWidth();
	self.h = Turbine.UI.Display:GetHeight();
	self.version = version;

	self:LoadSettings();

	self:SetBackColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
	self.l = Turbine.UI.Control();
	self.l:SetParent(self);
	self.l:SetSize(size, size);
	self.l:SetBackColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
	self.l:SetBackColorBlendMode(Turbine.UI.BlendMode.Multiply);
	self.l:SetMouseVisible(false);
	self.r = Turbine.UI.Control();
	self.r:SetParent(self);
	self.r:SetSize(size, size);
	self.r:SetBackColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
	self.r:SetBackColorBlendMode(Turbine.UI.BlendMode.Multiply);
	self.r:SetMouseVisible(false);
	self:Orient();

	self:SetWantsKeyEvents(true);

	self.threatTime = 5;
	self.powerFeed = 1400;
	self.player = Turbine.Gameplay.LocalPlayer.GetInstance();
	self.class = self.player:GetClass();
	self.effects = self.player:GetEffects();
	self.maxMorale = 1;
	self.maxPower = 1;
	self.inCombat = self.player:IsInCombat();

	self.memberOptions = { };
	self.memberMenu = Turbine.UI.ContextMenu();
	local items = self.memberMenu:GetItems();
	local f = function(self, args)
		self.parent.memberSelect.showPet = self.index;
		self.parent:PartyMemberChanged();
	end;
	for i = 1, #petOptions, 1 do
		self.memberOptions[i] = Turbine.UI.MenuItem(petOptions[i]);
		self.memberOptions[i].parent = self;
		self.memberOptions[i].index = i;
		self.memberOptions[i].Click = f;
		items:Add(self.memberOptions[i]);
	end

	self.tabOptions = { };
	self.tabMenu = Turbine.UI.ContextMenu();
	local items = self.tabMenu:GetItems();
	local l, o;
	local f = function(self, args)
		self.parent:Orient(self.layout, self.orient);
	end;
	self.tabOptions.options = { };
	for i = 1, #orientOptions, 1 do
		l = layoutOptions[i];
		o = orientOptions[i];
		self.tabOptions.options[i] = Turbine.UI.MenuItem(layoutName[l] .. " " .. orientName[o]);
		self.tabOptions.options[i].parent = self;
		self.tabOptions.options[i].layout = l;
		self.tabOptions.options[i].orient = o;
		self.tabOptions.options[i].Click = f;
		items:Add(self.tabOptions.options[i]);
	end

	self.players = { };
	self.actors = { };
	self.partyHandle = { };
	self.fellowtabs = { };
	self.fellowship = { };
	self.grid = { };

	self.options = PartyFrameOptions(self);

	self:UpdateParty();
	moebius92.Handlers.AddEventHandler(self.player, "InCombatChanged", function(sender, args) self:CombatChange(); end);
	moebius92.Handlers.AddEventHandler(self.player, "PartyChanged", function(sender, args) self:UpdateParty(); end);
	moebius92.Handlers.AddChatHandler(function(m) self:ChatMessage(m) end, Turbine.ChatType.Standard, nil);

	self:PetChanged();
	moebius92.Handlers.AddEventHandler(self.player, "PetChanged", function(sender, args) self:PetChanged(); end);
	moebius92.Handlers.AddEventHandler(self.effects, "EffectAdded", function(sender, args) self:LogEffect(self.effects:Get(args.Index)); end);
end

function PartyFrameWindow:MaxOffset()
	if self.majorX < 0 then
		self.maxOffset = math.max(0, math.floor(self.settings.x / -self.majorX));
	elseif self.majorX > 0 then
		self.maxOffset = math.max(0, math.floor((self.w - self.settings.x) / self.majorX));
	elseif self.majorY < 0 then
		self.maxOffset = math.max(0, math.floor(self.settings.y / -self.majorY));
	elseif self.majorY > 0 then
		self.maxOffset = math.max(0, math.floor((self.h - self.settings.y) / self.majorY));
	end
end

function PartyFrameWindow:Orient(layout, orient)
	local delta = 0;

	if layout ~= nil and self.settings.layout ~= layout then
		self.settings.layout = layout;
		self.settings.orient = orient;
		self.fellowship = { };
		self.grid = { };
		for _, v in pairs(self.actors) do
			v:Layout(self.settings.layout);
		end
		delta = 2;

		if layout == 1 then
			for _, v in pairs(self.players) do
				if v.showPet ~= 1 then
					v.showPet = 2;
				end
			end
		else
			local x = 1;
			if self.settings.orient <= 2 then
				if self.settings.x < (self.w / 2) then
					x = 5;
				else
					x = 4;
				end
			else
				if self.settings.y < (self.h / 2) then
					x = 2;
				else
					x = 3;
				end
			end
			for _, v in pairs(self.players) do
				if v.showPet ~= 1 then
					v.showPet = x;
				end
			end
		end
	elseif orient ~= nil and self.settings.orient ~= orient then
		self.settings.orient = orient;
		delta = 1;
	end

	self.dX = 0;
	self.dY = 0;
	self.majorX = 0;
	self.majorY = 0;
	self.minorX = 0;
	self.minorY = 0;

	if self.settings.layout == 1 then
		if self.settings.orient == 1 then
			self.dY = self.size;
			self.majorY = -self.size;
			self.minorX = self.size * 11 + self.width, self.size;
		else
			self.dY = -self.size;
			self.majorY = self.size;
			self.minorX = self.size * 11 + self.width, self.size;
		end
	else
		if self.settings.orient == 1 then
			self.dY = self.size * 3;
			self.majorY = -self.size * 3;
			self.minorX = self.size * 5;
		elseif self.settings.orient == 2 then
			self.dY = -self.size;
			self.majorY = self.size * 3;
			self.minorX = self.size * 5;
		elseif self.settings.orient == 3 then
			self.dX = -self.size;
			self.majorX = self.size * 5;
			self.minorY = self.size * 3;
		else
			self.dX = self.size * 5;
			self.majorX = -self.size * 5;
			self.minorY = self.size * 3;
		end
	end

	if self.settings.orient == 1 then
		self.l:SetBackground("moebius92/PartyFrame/ll.tga");
		self.r:SetBackground("moebius92/PartyFrame/lr.tga");
	elseif self.settings.orient == 2 then
		self.l:SetBackground("moebius92/PartyFrame/ul.tga");
		self.r:SetBackground("moebius92/PartyFrame/ur.tga");
	elseif self.settings.orient == 3 then
		self.l:SetBackground("moebius92/PartyFrame/ul.tga");
		self.r:SetBackground("moebius92/PartyFrame/ll.tga");
	else
		self.l:SetBackground("moebius92/PartyFrame/ur.tga");
		self.r:SetBackground("moebius92/PartyFrame/lr.tga");
	end
	local w = math.max(self.size, self.minorX);
	local h = math.max(self.size, self.minorY);
	self:SetSize(w, h);
	self.l:SetPosition(0, 0);
	self.r:SetPosition(math.max(0, self.minorX - self.size), math.max(0, self.minorY - self.size));
	if (self.settings.x + self.dX) >= self.w then
		self.settings.x = (self.w - 1) - self.dX;
		delta = math.max(1, delta);
	elseif (self.settings.x + self.dX + w) <= 0 then
		self.settings.x = 1 - (self.dX + w);
		delta = math.max(1, delta);
	end
	if (self.settings.y + self.dY) >= self.h then
		self.settings.y = (self.h - 1) - self.dY;
		delta = math.max(1, delta);
	elseif (self.settings.y + self.dY + h) <= 0 then
		self.settings.y = 1 - (self.dY + h);
		delta = math.max(1, delta);
	end

	self:MaxOffset();

	if delta > 0 then
		self:SaveSettings();
		if delta > 1 then
			self:PartyMemberChanged();
		else
			self:GridLayout();
		end
	end
end

function PartyFrameWindow:MouseDown(args)
	if args.Button == Turbine.UI.MouseButton.Left then
		self.dragStartX = args.X;
		self.dragStartY = args.Y;
		self.dragging = true;
	end
end

function PartyFrameWindow:MouseMove(args)
	if self.dragging then
		self.settings.x = self.settings.x + (args.X - self.dragStartX);
		self.settings.y = self.settings.y + (args.Y - self.dragStartY);

		if self.settings.layout == 1 then
			if self.settings.x < 0 then
				self.settings.x = 0;
			elseif self.settings.x > (self.w - (self.size * 11 + self.width)) then
				self.settings.x = self.w - (self.size * 11 + self.width);
			end
		end
		self:Layout();
	end
end

function PartyFrameWindow:MouseUp(args)
	if args.Button == Turbine.UI.MouseButton.Left and self.dragging then
		self.dragging = false;
		self:MaxOffset();
		self:SaveSettings();
	end
end

function PartyFrameWindow:MouseClick(args)
	if args.Button == Turbine.UI.MouseButton.Right then
		for i = 1, #self.tabOptions.options, 1 do
			self.tabOptions.options[i]:SetChecked(self.settings.layout == self.tabOptions.options[i].layout and self.settings.orient == self.tabOptions.options[i].orient);
		end
		self.tabMenu:ShowMenu();
	end
end

function PartyFrameWindow:LogEffect(effect)
	local debuff = effect:IsDebuff();
	local icon = effect:GetIcon();
	local name = effect:GetName();

	if debuff then
		if self.options.logDebuffs.index == 1 then
			return;
		elseif self.options.logDebuffs.index < 4 then
			if effectSlot["D/" .. icon .. "/" .. name] == nil and effectSlot["D/" .. name] ~= nil then
			elseif EffectSlot(effectSlot, debuff, icon, name, effect:GetCategory(), effect:IsCurable()) ~= nil then
				return;
			elseif self.options.logDebuffs.index == 2 then
				return;
			end
		end
	else
		if self.options.logBuffs.index == 1 then
			return;
		elseif self.options.logBuffs.index < 4 then
			if effectSlot["B/" .. icon .. "/" .. name] == nil and effectSlot["B/" .. name] ~= nil then
			elseif EffectSlot(effectSlot, debuff, icon, name, effect:GetCategory(), effect:IsCurable()) ~= nil then
				return;
			elseif self.options.logBuffs.index == 2 then
				return;
			end
		end
	end

	for _, v in pairs(self.options.slots[0]) do
		if v.name == name and v.debuff == debuff and v.icon == icon then
			return;
		end
	end

	table.insert(self.options.slots[0], { name = name, debuff = debuff, icon = icon });
	table.sort(self.options.slots[0], function(a, b) return a.name < b.name; end);
	self.options:SizeChanged();
end

function PartyFrameWindow:PetChanged()
	if self.pet ~= nil then
		moebius92.Handlers.RemoveEventHandler(self.petEffects, "EffectAdded", self.petHandler);
		self.pet = nil;
	end

	self.pet = self.player:GetPet();

	if self.pet ~= nil then
		self.petEffects = self.pet:GetEffects();
		self.petHandler = moebius92.Handlers.AddEventHandler(self.petEffects, "EffectAdded", function(sender, args) self:LogEffect(self.petEffects:Get(args.Index)); end);
	end
end

function PartyFrameWindow:LoadSettings()
	self.settings = Turbine.PluginData.Load(Turbine.DataScope.Character, "PartyFrame");
	if type(self.settings) == "string" then
		self.settings = loadstring("return " .. self.settings .. ";")();
	end

	if self.settings == nil or self.settings.version ~= self.version then
		self.settings = { };
		self.settings.x = self.w - (self.size * 11 + self.width);
		self.settings.y = self.h / 2 - self.size;
		self.settings.orient = 1;
		self.settings.layout = 1;
	end
	self.settings.version = nil;
end

function PartyFrameWindow:SaveSettings()
	self.settings.version = self.version;
	Turbine.PluginData.Save(Turbine.DataScope.Character, "PartyFrame", serialize(self.settings));
	self.settings.version = nil;
end

function PartyFrameWindow:KeyDown(args)
	if args.Action == 0x1000007B then
		if self.move then
			if self.dragging then
				self.dragging = false;
				self:SaveSettings();
			end
			for _, v in pairs(self.fellowtabs) do
				v.dragging = false;
			end
			for k, v in pairs(self.players) do
				v.dragging = false;
				v.select:SetMouseVisible(true);
			end
			self.move = false;
		else
			for _, v in pairs(self.players) do
				v.select:SetMouseVisible(false);
			end
			self.move = true;
		end
		self:Layout();
	elseif args.Action == 0x100000B3 then
		for _, v in pairs(self.actors) do
			v:SetVisible(self.hide);
		end
		self.hide = not self.hide;
	end
	self:SetVisible(not self.hide and self.move);
end

function PartyFrameWindow:ChatMessage(m)
	local i = string.find(m, " ");

	if i ~= nil then
		local name = string.sub(m, 1, i - 1);

		m = string.sub(m, i + 1, -1);
		if string.sub(m, 1, 14) == "rolled (100): " then
			if self.players[name] ~= nil then
				local t = { };
				local i, j, k;
				local state, w;

				table.insert(self.players[name].rolls, tonumber(string.sub(m, 15, -1)));

				for k, v in pairs(self.players) do
					table.insert(t, { ["rolls"] = v.rolls, ["name"] = k });
				end
				table.sort(t, 
function(a, b)
	local n = math.min(#a.rolls, #b.rolls);
	local i;

	for i = 1, math.min(#a.rolls, #b.rolls), 1 do
		if a.rolls[i] > b.rolls[i] then
			return true;
		elseif a.rolls[i] < b.rolls[i] then
			return false;
		end
	end

	return #a.rolls > #b.rolls;
end
				          );

				k = 0;
				for i = 1, #t, 1 do
					if #t[i].rolls == 0 then
						break;
					end

					state = 0;
					if i == 1 then
						state = 2
					else
						for j = 1, #t[i].rolls, 1 do
							if t[i - 1].rolls[j] > t[i].rolls[j] then
								state = 2;
								break;
							end
						end

						if state == 0 then
							if #t[i - 1].rolls > #t[i].rolls then
								state = 1;
							end
						end
					end

					if state == 2 then
						k = k + 1;
						w = k;
					elseif state == 1 then
						w = k .. "?";
					else
						w = k;
					end

					if self.players[t[i].name].rollOrder ~= w then
						self.players[t[i].name].rollOrder = w;
						self.players[t[i].name]:UpdateName();
					end
				end
			end
		end
	end
end

function PartyFrameWindow:AddMember(member, name)
	if self.players[name] == nil then
		self.players[name] = PartyMemberWindow(self, member, self.size, self.width, nil);
		if self.settings.layout == 1 then
			self.players[name].showPet = 2;
		else
			if self.settings.orient <= 2 then
				if self.settings.x < (self.w / 2) then
					self.players[name].showPet = 5;
				else
					self.players[name].showPet = 4;
				end
			else
				if self.settings.y < (self.h / 2) then
					self.players[name].showPet = 2;
				else
					self.players[name].showPet = 3;
				end
			end
		end
	end

	self.players[name].id = name;

	if not self.players[name].exists then
		self.players[name].exists = true;
		self.players[name]:SetVisible(true);
		self.players[name]:UpdateReadyState();
		self.players[name].select:SetMouseVisible(not self.move);
		table.insert(self.actors, self.players[name]);

		local pet = member:GetPet();

		if self.players[name].showPet == 1 then
			pet = nil;
		end

		if self.players[name].pet == nil or self.players[name].pet.actor ~= pet then
			if self.players[name].pet ~= nil then
				self.players[name].pet:RemoveHandlers();
				self.players[name].pet:SetVisible(false);
				self.players[name].pet:SetParent(nil);
				self.players[name].pet = nil;
			end

			if pet ~= nil then
				self.players[name].pet = PartyMemberWindow(self, pet, self.size, self.width, self.players[name].class);
			end
		end

		if self.players[name].pet ~= nil then
			self.players[name].pet:SetVisible(true);
			table.insert(self.actors, self.players[name].pet);
		end
	end
end

function sign(x)
	if x > 0 then
		return 1;
	elseif x < 0 then
		return -1;
	end
	return 0;
end

function PartyFrameWindow:MoveMember(id, dx, dy)
	local x = self.players[id].indexX;
	local y = self.players[id].indexY;

	if self.settings.layout == 1 then
		if dx == 0 then
			if dy < 0 then
				if y > 1 then
					self.fellowship[x][y] = self.fellowship[x][y - 1];
					self.fellowship[x][y - 1] = id;
				else
					table.remove(self.fellowship[x], y);
					x = x - orientDFellowship[self.settings.orient];
					if self.fellowship[x] ~= nil then
						table.insert(self.fellowship[x], id);
					elseif x == 0 then
						table.insert(self.fellowship, 1, { [1] = id });
					else
						self.fellowship[x] = { [1] = id };
					end
				end
				self:GridLayout();
				return true;
			elseif dy > 0 then
				if y < #self.fellowship[x] then
					self.fellowship[x][y] = self.fellowship[x][y + 1];
					self.fellowship[x][y + 1] = id;
					self:GridLayout();
					return true;
				else
					table.remove(self.fellowship[x], y);
					x = x + orientDFellowship[self.settings.orient];
					if self.fellowship[x] ~= nil then
						table.insert(self.fellowship[x], 1, id);
					elseif x == 0 then
						table.insert(self.fellowship, 1, { [1] = id });
					else
						self.fellowship[x] = { [1] = id };
					end
				end
				self:GridLayout();
				return true;
			end
		else
			if dy < 0 then
				table.remove(self.fellowship[x], y);
				if orientDFellowship[self.settings.orient] > 0 then
					table.insert(self.fellowship, x, { [1] = id });
				else
					table.insert(self.fellowship, x + 1, { [1] = id });
				end
				self:GridLayout();
				return true;
			elseif dy > 0 then
				table.remove(self.fellowship[x], y);
				if orientDFellowship[self.settings.orient] < 0 then
					table.insert(self.fellowship, x, { [1] = id });
				else
					table.insert(self.fellowship, x + 1, { [1] = id });
				end
				self:GridLayout();
				return true;
			end
		end
	else
		local minor = sign(dx) * sign(self.majorX) + sign(dy) * sign(self.majorY);
		local major = sign(dx) * sign(self.minorX) + sign(dy) * sign(self.minorY);

		if major ~= 0 then
			if self.grid[x][y].offset + major >= 0 then
				if self.grid[x][y + major] ~= nil and self.grid[x][y + major].offset == (self.grid[x][y].offset + major) then
					self.grid[x][y].name = self.grid[x][y + major].name;
					self.grid[x][y + major].name = id;
				else
					self.grid[x][y].offset = self.grid[x][y].offset + major;
				end
				self:GridLayout();
				return true;
			end
		elseif minor ~= 0 then
			if self.grid[x + minor] ~= nil then
				for _, v in ipairs(self.grid[x + minor]) do
					if v.offset == self.grid[x][y].offset then
						self.grid[x][y].name = v.name;
						v.name = id;
						self:GridLayout();
						return true;
					end
				end

				table.insert(self.grid[x + minor], { offset = self.grid[x][y].offset, name = id });
				table.sort(self.grid[x + minor], function(a, b) return a.offset < b.offset; end);
				table.remove(self.grid[x], y);
				self:GridLayout();
				return true;
			else
				self.grid[x + minor] = { [1] = table.remove(self.grid[x], y) };
				self:GridLayout();
				return true;
			end
		end
	end
	return false;
end

function PartyFrameWindow:UpdateParty()
	for k, v in pairs(self.partyHandle) do
		moebius92.Handlers.RemoveEventHandler(self.party, k, v);
	end
	self.partyHandle = { };

	for k, v in pairs(self.actors) do
		v:RemoveHandlers();
		v:SetVisible(false);
		v:SetParent(nil);
	end
	self.players = { };
	self.actors = { };

	self.party = self.player:GetParty();
	if self.party ~= nil then
		self.partyHandle["MemberAdded"] = moebius92.Handlers.AddEventHandler(self.party, "MemberAdded", function(sender, args) self:PartyMemberChanged(); end);
		self.partyHandle["MemberRemoved"] = moebius92.Handlers.AddEventHandler(self.party, "MemberRemoved", function(sender, args) self:PartyMemberChanged(); end);
		self.partyHandle["LeaderChanged"] = moebius92.Handlers.AddEventHandler(self.party, "LeaderChanged", function(sender, args) self:PartyMemberChanged(); end);
	end

	self:PartyMemberChanged();
end

function PartyFrameWindow:PartyMemberChanged()
	local t = { };
	local n;

	self.actors = { };
	for k, v in pairs(self.players) do
		v.exists = false;
		v.leader = false;
	end

	if self.party ~= nil and self.party:GetMemberCount() >= 1 then
		local n = self.party:GetMemberCount();
		local l = self.party:GetLeader();
		local name;

		for i = 1, n, 1 do
			local member = self.party:GetMember(i);
			name = member:GetName();

			self:AddMember(member, name);
			self.players[name].index = i;
			if l == member then
				self.players[name].leader = true;
				self.players[name]:UpdateReadyState();
			end
		end

		name = self.player:GetName();
		if self.players[name] == nil then
			self:AddMember(self.player, name);
			self.players[name].index = 0;
		end
	end

	for k, v in pairs(self.players) do
		if not v.exists then
			v:RemoveHandlers();
			v:SetVisible(false);
			v:SetParent(nil);
			if v.pet ~= nil then
				v.pet:RemoveHandlers();
				v.pet:SetVisible(false);
				v.pet:SetParent(nil);
			end
			self.players[k] = nil;
		else
			v.free = true;
		end
	end

	self:UpdateMorale();
	self:UpdatePower();
	for _, v in pairs(self.actors) do
		v:UpdateMorale();
		v:UpdatePower();
	end

	if self.settings.layout == 1 then
		for k, v in ipairs(self.fellowship) do
			t = { };
			for _, v1 in ipairs(v) do
				if self.players[v1] ~= nil and self.players[v1].free then
					self.players[v1].free = nil;
					table.insert(t, v1);
				end
			end

			self.fellowship[k] = t;
		end

		for i = #self.fellowship, 1, -1 do
			if #self.fellowship[i] == 0 then
				table.remove(self.fellowship, i);
			end
		end

		t = { };
		for k, v in pairs(self.players) do
			if v.free then
				v.free = nil;
				table.insert(t, { name = k, index = v.index });
			end
		end
		table.sort(t, function(a, b) return a.index < b.index; end);

		n = #self.fellowship;
		if n == 0 then
			self.fellowship[1] = { };
			n = 1;
		end

		for _, v in ipairs(t) do
			table.insert(self.fellowship[n], v.name);
		end
	else
		self.fellowship = { };

		for k, v in pairs(self.grid) do
			n = 0;
			t = { };
			for _, v1 in ipairs(v) do
				if self.players[v1.name] ~= nil and self.players[v1.name].free then
					self.players[v1.name].free = nil;
					table.insert(t, { offset = v1.offset - n, name = v1.name });
				else
					n = n + 1;
				end
			end
			self.grid[k] = t;
		end

		t = { };
		for k, v in pairs(self.players) do
			if v.free then
				v.free = nil;
				table.insert(t, { name = k, index = v.index });
			end
		end
		table.sort(t, function(a, b) return a.index < b.index; end);

		n = 0;
		for _, v in ipairs(t) do
			local offset;

			while self.grid[n] ~= nil and #self.grid[n] > 0 and self.grid[n][#self.grid[n]].offset >= self.maxOffset do
				if self.settings.orient <= 2 then
					n = n + ((self.settings.y < (self.h / 2)) and -1 or 1);
				else
					n = n + ((self.settings.x < (self.w / 2)) and -1 or 1);
				end
			end

			if self.grid[n] == nil then
				self.grid[n] = { };
			end

			if #self.grid[n] == 0 then
				offset = 0;
			else
				offset = self.grid[n][#self.grid[n]].offset + 1;
			end

			table.insert(self.grid[n], { offset = offset, name = v.name});
		end
	end

	self:GridLayout();
end

function PartyFrameWindow:GridLayout()
	if self.settings.layout == 1 then
		for i = #self.fellowship, 1, -1 do
			if #self.fellowship[i] == 0 then
				table.remove(self.fellowship, i);
			end
		end

		self.grid = { };
		self.grid[0] = { };
		n = 0;
		for k, v in ipairs(self.fellowship) do
			local l = 0;

			if self.fellowtabs[k] == nil then
				self.fellowtabs[k] = PartyFrameFTab(self, k);
			end
			self.fellowtabs[k].s = n;

			if self.settings.orient == 1 then
				for i = #v, 1, -1 do
					l = l + 1;
					if self.players[v[i]].pet then
						n = n + 1;
						l = l + 1;
					end
					table.insert(self.grid[0], { offset = n, name = v[i] });
					self.players[v[i]].indexX = k;
					self.players[v[i]].indexY = i;
					n = n + 1;
				end
			else
				for i = 1, #v, 1 do
					l = l + 1;
					table.insert(self.grid[0], { offset = n, name = v[i] });
					self.players[v[i]].indexX = k;
					self.players[v[i]].indexY = i;
					if self.players[v[i]].pet then
						n = n + 1;
						l = l + 1;
					end
					n = n + 1;
				end
			end

			self.fellowtabs[k].l = l;

			n = n + 1;
		end
	else
		for k, v in pairs(self.grid) do
			for k1, v1 in ipairs(v) do
				self.players[v1.name].indexX = k;
				self.players[v1.name].indexY = k1;
			end
		end
	end

	for k, v in pairs(self.grid) do
		if #v == 0 then
			self.grid[k] = nil;
		end
	end

	self:Layout();
end

petDX = { };
petDX[2] = 0;
petDX[3] = 0;
petDX[4] = -1;
petDX[5] = 1;

petDY = { };
petDY[2] = 1;
petDY[3] = -1;
petDY[4] = 0;
petDY[5] = 0;

function PartyFrameWindow:Layout()
	local h, w = self:GetSize();

	self:SetPosition(self.settings.x + self.dX, self.settings.y + self.dY);

	if self.settings.layout == 1 then
		local o = self.settings.x < (self.w / 2);

		for i = 1, #self.fellowship, 1 do
			self.fellowtabs[i]:Orient(o);
			if self.settings.orient == 1 then
				self.fellowtabs[i]:SetPosition(self.settings.x + (o and self.minorX or -self.size), self.settings.y + (self.fellowtabs[i].s + self.fellowtabs[i].l - 1) * self.majorY);
			else
				self.fellowtabs[i]:SetPosition(self.settings.x + (o and self.minorX or -self.size), self.settings.y + self.fellowtabs[i].s * self.majorY);
			end
		end
		for i = #self.fellowship + 1, #self.fellowtabs, 1 do
			self.fellowtabs[i]:SetVisible(false);
		end

		if not self.move and self.grid[0] ~= nil and #self.grid[0] > 0 then
			local n = self.grid[0][#self.grid[0]].offset;

			if self.settings.orient == 1 then
				if n >= self.maxOffset then
					for _, v in pairs(self.grid[0]) do
						self.players[v.name]:SetPosition(self.settings.x, 0 + (v.offset - n) * self.majorY);
						if self.players[v.name].pet ~= nil then
							self.players[v.name].pet:SetPosition(self.settings.x, 0 + (v.offset - 1 - n) * self.majorY);
						end
					end
					return;
				end
			else
				if self.grid[0][#self.grid[0]].pet ~= nil then
					n = n + 1;
				end

				if n >= self.maxOffset then
					for _, v in pairs(self.grid[0]) do
						self.players[v.name]:SetPosition(self.settings.x, self.h + (v.offset - n) * self.majorY);
						if self.players[v.name].pet ~= nil then
							self.players[v.name].pet:SetPosition(self.settings.x, self.h + (v.offset + 1 - n) * self.majorY);
						end
					end
					return;
				end
			end
		end
	else
		for i = 1, #self.fellowtabs, 1 do
			self.fellowtabs[i]:SetVisible(false);
		end
	end

	for k, v in pairs(self.grid) do
		for _, v1 in pairs(v) do
			self.players[v1.name]:SetPosition(self.settings.x + k * self.minorX + v1.offset * self.majorX,
											  self.settings.y + k * self.minorY + v1.offset * self.majorY);
			if self.players[v1.name].pet ~= nil then
				self.players[v1.name].pet:SetPosition(self.settings.x + k * self.minorX + v1.offset * self.majorX + petDX[self.players[v1.name].showPet] * math.abs(self.majorX + self.minorX),
													  self.settings.y + k * self.minorY + v1.offset * self.majorY + petDY[self.players[v1.name].showPet] * math.abs(self.majorY + self.minorY));
			end
		end
	end
end

function PartyFrameWindow:UpdateMorale()
	local m = 1;

	for k, v in pairs(self.actors) do
		m = math.max(m, v.maxMorale);
	end

	if m ~= self.maxMorale then
		self.maxMorale = m;
		for k, v in pairs(self.actors) do
			v:UpdateMorale();
		end
	end
end

function PartyFrameWindow:UpdatePower()
	local m = 1;

	for k, v in pairs(self.actors) do
		m = math.max(m, v.maxPower);
	end

	if m ~= self.maxPower then
		self.maxPower = m;
		for k, v in pairs(self.actors) do
			v:UpdatePower();
		end
	end
end

function PartyFrameWindow:CombatChange()
	self.inCombat = self.player:IsInCombat();
	for k, v in pairs(self.actors) do
		v.damageHead = nil;
		v.damageTail = nil;
		v.damage = 0;
		v.threat = 0;
		v:UpdateMorale();
	end
end

PartyFrameOptionsButton = class(Turbine.UI.Control);

function PartyFrameOptionsButton:Constructor(parent, index, missing)
	Turbine.UI.Control.Constructor(self);

	self.parent = parent;
	self.index = index;
	self.highlight = false;
	self.missing = missing;

	self:SetParent(parent);
	self:SetSize(parent.window.size - 2, parent.window.size - 2);
	self:Colorize();
end

function PartyFrameOptionsButton:SetMissing(m)
	if m ~= self.missing then
		if m then
			local t = self.parent.slots[self.index];

			self.parent.slots[self.index] = { };
			if #t > 0 then
				self.parent.slots[self.index][1] = t;
			end
		else
			local t = self.parent.slots[self.index];

			self.parent.slots[self.index] = { };
			for _, v in pairs(t) do
				for _, v1 in pairs(v) do
					table.insert(self.parent.slots[self.index], v1);
				end
			end
			table.sort(self.parent.slots[self.index], function(a, b) return a.name < b.name; end);
		end

		self.missing = m;
		self:Colorize();
		self.parent:Changed();
	end
end

function PartyFrameOptionsButton:Colorize()
	local name = self.parent.setupName:GetText();
	local c = 0.5;

	if self.parent.select == self.index then
		c = 1.0;
	end

	if self.missing then
		if name and self.parent.setupSaved[name] and self.parent.setupSaved[name][self.index] then
			self:SetBackColor(Turbine.UI.Color(1.0, c, c / 2, 0.0));
		else
			self:SetBackColor(Turbine.UI.Color(1.0, c, 0.0, 0.0));
		end
	else
		if name and self.parent.setupSaved[name] and self.parent.setupSaved[name][self.index] then
			self:SetBackColor(Turbine.UI.Color(1.0, c / 2, c, c / 2));
		else
			self:SetBackColor(Turbine.UI.Color(1.0, c, c, c));
		end
	end
end

function PartyFrameOptionsButton:MouseClick(args)
	self.parent:Select(self.index);
end

function PartyFrameOptionsButton:MouseEnter(args)
	if self.parent.dragSource ~= nil and self.parent.dragTime == Turbine.Engine.GetGameTime() then
		local x, y = self.parent:GetIndices(self.parent.dragSource.index);
		local t;

		if y == nil then
			t = self.parent.slots[self.parent.select][x];
			if self.parent.select ~= 0 then
				table.remove(self.parent.slots[self.parent.select], x);
			end
		else
			t = self.parent.slots[self.parent.select][x][y];
			if self.parent.select ~= 0 then
				table.remove(self.parent.slots[self.parent.select][x], y);
				if #self.parent.slots[self.parent.select][x] == 0 then
					table.remove(self.parent.slots[self.parent.select], x);
				end
			end
		end

		if self.missing then
			table.insert(self.parent.slots[self.index], { [1] = t });
		else
			table.insert(self.parent.slots[self.index], t);
			table.sort(self.parent.slots[self.index], function(a, b) return (not a.name and "" or a.name) < (not b.name and "" or b.name); end);
		end

		self.parent:Changed();
	end
end

PartyFrameOptionsDropDown = class(Turbine.UI.Lotro.Button)

function PartyFrameOptionsDropDown:Constructor(parent, index, func, ...)
	Turbine.UI.Lotro.Button.Constructor(self);

	self.parent = parent;
	self.label = arg;
	self.func = func;
	self.items = { };

	self:SetParent(parent);
	self.dropdown = Turbine.UI.ContextMenu();

	local items = self.dropdown:GetItems();
	for k, v in ipairs(self.label) do
		self.items[k] = PartyFrameOptionsDDItem(v, self, k);
		items:Add(self.items[k]);
	end

	self:SetIndex(index);
end

function PartyFrameOptionsDropDown:SetIndex(index, click)
	for k, v in ipairs(self.items) do
		v:SetChecked(k == index);
	end
	self:SetText(self.label[index]);
	self.index = index;
	if click and self.func ~= nil then
		self.func(self, self.index);
	end
end

function PartyFrameOptionsDropDown:MouseClick(args)
	self.dropdown:ShowMenuAt(self:PointToScreen(0, self:GetHeight()));
end

PartyFrameOptionsDDItem = class(Turbine.UI.MenuItem)

function PartyFrameOptionsDDItem:Constructor(label, parent, index)
	Turbine.UI.MenuItem.Constructor(self, label);

	self.parent = parent;
	self.index = index;
end

function PartyFrameOptionsDDItem:Click(args)
	self.parent:SetIndex(self.index, true);
end

PartyFrameOptionsEffectText = class(Turbine.UI.Lotro.TextBox);

function PartyFrameOptionsEffectText:Constructor(parent)
	Turbine.UI.Lotro.TextBox.Constructor(self);

	self.parent = parent;
	self:SetParent(parent);
	self:SetMultiline(false);
	self:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
	self:SetFont(Turbine.UI.Lotro.Font.Verdana16);
	self:SetWantsKeyEvents(true);
end

function PartyFrameOptionsEffectText:KeyDown(args)
	if self:HasFocus() and args.Action == 162 then
		local name = self:GetText();
		local option = self.parent.parent;
		local x, y = option:GetIndices(self.parent.index);

		self.parent:Focus();
		if option.slotsButton[option.select].missing then
			if x == nil then
				table.insert(option.slots[option.select], { [1] = { name = name, debuff = false } });
			else
				option.slots[option.select][x][y].name = name;
			end
		else
			if x == nil then
				table.insert(option.slots[option.select], { name = name, debuff = false });
			else
				option.slots[option.select][x].name = name;
			end
			table.sort(option.slots[option.select], function(a, b) return (not a.name and "" or a.name) < (not b.name and "" or b.name); end);
		end
		option:Changed();
	end
end

PartyFrameOptionsEffect = class(Turbine.UI.Control)

function PartyFrameOptionsEffect:Constructor(parent, index, icon, debuff, name, ro)
	Turbine.UI.Control.Constructor(self);

	self.parent = parent;
	self.index = index;
	self.dragging = false;
	self:SetParent(parent);
	self:SetMouseVisible(true);

	self.icon = Turbine.UI.Control();
	self.icon:SetParent(self);
	self.icon:SetMouseVisible(false);
	self.icon:SetBackColor(Turbine.UI.Color(1.0, 0.0, 0.0, 0.0));
	self.icon:SetPosition(1, 1);
	self.textbox = PartyFrameOptionsEffectText(self);
	self:SetIcon(icon);
	self:SetDebuff(debuff);
	self:SetName(name, ro);
end

function PartyFrameOptionsEffect:MouseDown(args)
	if args.Button == Turbine.UI.MouseButton.Left then
		local x, y = Turbine.UI.Display.GetMousePosition();

		self.parent.dragWindow:SetBackground(self.image);
		if self.image ~= nil then
			self.parent.dragWindow:SetBackColorBlendMode(Turbine.UI.BlendMode.Screen);
		else
			self.parent.dragWindow:SetBackColorBlendMode(Turbine.UI.BlendMode.Normal);
		end
		self.parent.dragWindow:Activate();
		self.parent.dragWindow:SetZOrder(5000);
		self.parent.dragWindow:SetVisible(true);
		self.parent.dragWindow:SetPosition(x - 16, y - 16);
		self.dragging = true;
	end
end

function PartyFrameOptionsEffect:MouseMove(args)
	if self.dragging then
		local x, y = Turbine.UI.Display.GetMousePosition();
		self.parent.dragWindow:SetPosition(x - 16, y - 16);
	end
end

function PartyFrameOptionsEffect:MouseEnter(args)
	if self.parent.dragSource ~= nil and self.parent.dragTime == Turbine.Engine.GetGameTime() and self.parent.select ~= 0 then
		local tx, ty = self.parent:GetIndices(self.index);
		local sx, sy = self.parent:GetIndices(self.parent.dragSource.index);
		local x, y = self:GetMousePosition();
		local d = (y > (self.parent.window.size / 2)) and 1 or 0;

		if sx == nil then
			if tx == nil then
				if self.parent.slotsButton[self.parent.select].missing then
					table.insert(self.parent.slots[self.parent.select], { [1] = { debuff = false } });
				else
					table.insert(self.parent.slots[self.parent.select], { debuff = false });
				end
			elseif ty == nil then
				table.insert(self.parent.slots[self.parent.select], tx + d, { debuff = false });
			else
				table.insert(self.parent.slots[self.parent.select][tx], ty + d, { debuff = false });
			end
			self.parent:Changed();
		elseif self ~= self.parent.dragSource and tx ~= nil and self.parent.slotsButton[self.parent.select].missing then
			if sx ~= tx or sy < ty then
				table.insert(self.parent.slots[self.parent.select][tx], ty + d, self.parent.slots[self.parent.select][sx][sy]);
				table.remove(self.parent.slots[self.parent.select][sx], sy);
				if #self.parent.slots[self.parent.select][sx] == 0 then
					table.remove(self.parent.slots[self.parent.select], sx);
				end
			else
				local t = self.parent.slots[self.parent.select][sx][sy];

				table.remove(self.parent.slots[self.parent.select][sx], sy);
				table.insert(self.parent.slots[self.parent.select][tx], ty + d, t);
				if #self.parent.slots[self.parent.select][sx] == 0 then
					table.remove(self.parent.slots[self.parent.select], sx);
				end
			end
			self.parent:Changed();
		end
	end
end

function PartyFrameOptionsEffect:MouseUp(args)
	if args.Button == Turbine.UI.MouseButton.Left then
		self.parent.dragWindow:SetVisible(false);
		self.dragging = false;
		self.parent.dragSource = self;
		self.parent.dragTime = Turbine.Engine.GetGameTime();
	end
end

function PartyFrameOptionsEffect:MouseClick(args)
	if args.Button == Turbine.UI.MouseButton.Right then
		local x, y = self.parent:GetIndices(self.index);

		self.parent.effectsMenuTarget = self;
		self.parent.effectsOptions.Debuff:SetChecked(self.debuff);
		self.parent.effectsIndex = { x = x, y = y };
		if self.parent.select == 0 then
			self.parent.effectsOptions.Cut:SetEnabled(true);
			self.parent.effectsOptions.Copy:SetEnabled(true);
			self.parent.effectsOptions.Paste:SetEnabled(false);
			self.parent.effectsOptions.Insert:SetEnabled(false);
			self.parent.effectsOptions.Delete:SetEnabled(true);
			self.parent.effectsOptions.Debuff:SetEnabled(false);
			self.parent.effectsOptions.Save:SetEnabled(true);
			self.parent.effectsOptions.NewGroup:SetEnabled(false);
		else
			self.parent.effectsOptions.Cut:SetEnabled(x ~= nil);
			self.parent.effectsOptions.Copy:SetEnabled(x ~= nil);
			self.parent.effectsOptions.Paste:SetEnabled(self.parent.effectsClipboard ~= nil);
			self.parent.effectsOptions.Insert:SetEnabled(true);
			self.parent.effectsOptions.Delete:SetEnabled(x ~= nil);
			self.parent.effectsOptions.Debuff:SetEnabled(x ~= nil);
			self.parent.effectsOptions.Save:SetEnabled(x ~= nil);
			self.parent.effectsOptions.NewGroup:SetEnabled(self.parent.slotsButton[self.parent.select].missing and x ~= nil and y > 1);
		end
		self.parent.effectsMenu:ShowMenu();
	end
end

function PartyFrameOptionsEffect:SetIcon(icon)
	local w, h = self:GetSize();

	self.icon:SetBackground(icon);
	self.image = icon;
	if icon == nil then
		self.icon:SetBackColorBlendMode(Turbine.UI.BlendMode.Normal);
	else
		self.icon:SetBackColorBlendMode(Turbine.UI.BlendMode.Screen);
	end
end

function PartyFrameOptionsEffect:SetName(name, ro)
	self.textbox:SetText(name);
	self.textbox:SetReadOnly(ro);
end

function PartyFrameOptionsEffect:SetDebuff(debuff)
	self.debuff = debuff;
	if debuff then
		self.textbox:SetForeColor(Turbine.UI.Color(1.0, 1.0, 0.0, 0.0));
	else
		self.textbox:SetForeColor(Turbine.UI.Color(1.0, 0.0, 1.0, 0.0));
	end
end

function PartyFrameOptionsEffect:SizeChanged(args)
	local w, h = self:GetSize();

	self.icon:SetSize(h - 2, h - 2);
	self.textbox:SetSize(w - (h + 2), h - 2);
	self.textbox:SetPosition(h + 1, 1);
end

PartyFrameOptionsGroup = class(Turbine.UI.Control);

function PartyFrameOptionsGroup:Constructor(parent, index)
	Turbine.UI.Control.Constructor(self);

	self.parent = parent;
	self.index = index;
	self.dragging = false;

	self:SetParent(parent);

	self.u = Turbine.UI.Control();
	self.u:SetParent(self);
	self.u:SetPosition(0, 0);
	self.u:SetSize(parent.window.size, parent.window.size);
	self.u:SetBackground("moebius92/PartyFrame/ul.tga");
	self.u:SetBlendMode(Turbine.UI.BlendMode.Overlay);
	self.u:SetMouseVisible(false);
	self.c = Turbine.UI.Control();
	self.c:SetParent(self);
	self.c:SetSize(parent.window.size, parent.window.size);
	self.c:SetBackColor(Turbine.UI.Color(1.0, 1.0, 1.0, 1.0));
	self.c:SetPosition(0, parent.window.size);
	self.c:SetMouseVisible(false);
	self.d = Turbine.UI.Control();
	self.d:SetParent(self);
	self.d:SetSize(parent.window.size, parent.window.size);
	self.d:SetBackground("moebius92/PartyFrame/ll.tga");
	self.d:SetBlendMode(Turbine.UI.BlendMode.Overlay);
	self.d:SetMouseVisible(false);
end

function PartyFrameOptionsGroup:Setup(l)
	if l > 0 then
		self:SetSize(self.parent.window.size, (self.parent.window.size + 2) * l - 2);

		if l == 1 then
			self.c:SetVisible(false);
			self.d:SetVisible(false);
		else
			self.d:SetPosition(0, (self.parent.window.size + 2) * (l - 1) - 2);
			self.d:SetVisible(true);
			self.c:SetSize(self.parent.window.size, (self.parent.window.size + 2) * (l - 2));
			self.c:SetVisible(true);
		end

		self:SetVisible(true);
	else
		self:SetVisible(false);
	end
end

function PartyFrameOptionsGroup:MouseDown(args)
	self.dragging = true;
	self.safeY = nil;
end

function PartyFrameOptionsGroup:MouseMove(args)
	if self.dragging then
		if not IsSafeY(self, args.Y) then
			return;
		end

		if args.Y < 0 then
			local i = self.index - 1;

			if self.parent.groups[i] ~= nil then
				local x = self.parent.slots[self.parent.select][self.index];

				self.parent.slots[self.parent.select][self.index] = self.parent.slots[self.parent.select][i];
				self.parent.slots[self.parent.select][i] = x;
				self.parent.groups[self.index] = self.parent.groups[i];
				self.parent.groups[self.index].index = self.index;
				self.parent.groups[i] = self;
				self.index = i;
				self.parent:Changed();
				self.safeY = math.huge;
			end
		elseif args.Y > self:GetHeight() then
			local i = self.index + 1;

			if self.parent.groups[i] ~= nil then
				local x = self.parent.slots[self.parent.select][self.index];

				self.parent.slots[self.parent.select][self.index] = self.parent.slots[self.parent.select][i];
				self.parent.slots[self.parent.select][i] = x;
				self.parent.groups[self.index] = self.parent.groups[i];
				self.parent.groups[self.index].index = self.index;
				self.parent.groups[i] = self;
				self.index = i;
				self.parent:Changed();
				self.safeY = -math.huge;
			end
		end
	end
end

function PartyFrameOptionsGroup:MouseUp(args)
	self.dragging = false;
end

PartyFrameOptions = class(Turbine.UI.Control);

function PartyFrameOptions:Constructor(window)
	Turbine.UI.Control.Constructor(self);

	self.window = window;
	self.select = 0;
	self.width = 100;
	self.effects = { };
	self.groups = { };

	self.effectsClipboard = nil;
	self.effectsIndex = { };
	self.effectsOptions = { };
	self.effectsMenu = Turbine.UI.ContextMenu();
	local items = self.effectsMenu:GetItems();
	self.effectsOptions.Cut = Turbine.UI.MenuItem("Cut");
	self.effectsOptions.Cut.parent = self;
	self.effectsOptions.Cut.Click = function(self, args)
		local x = self.parent.effectsIndex.x;
		local y = self.parent.effectsIndex.y;

		if y == nil then
			self.parent.effectsClipboard = self.parent.slots[self.parent.select][x];
			table.remove(self.parent.slots[self.parent.select], x);
		else
			self.parent.effectsClipboard = self.parent.slots[self.parent.select][x][y];
			table.remove(self.parent.slots[self.parent.select][x], y);
			if #self.parent.slots[self.parent.select][x] == 0 then
				table.remove(self.parent.slots[self.parent.select], x);
			end
		end
		self.parent:Changed();
	end;
	items:Add(self.effectsOptions.Cut);
	self.effectsOptions.Copy = Turbine.UI.MenuItem("Copy");
	self.effectsOptions.Copy.parent = self;
	self.effectsOptions.Copy.Click = function(self, args)
		local x = self.parent.effectsIndex.x;
		local y = self.parent.effectsIndex.y;
		local t;

		if y == nil then
			t = self.parent.slots[self.parent.select][x];
		else
			t = self.parent.slots[self.parent.select][x][y];
		end
		self.parent.effectsClipboard = { debuff = t.debuff, icon = t.icon, name = t.name };
	end;
	items:Add(self.effectsOptions.Copy);
	self.effectsOptions.Paste = Turbine.UI.MenuItem("Paste");
	self.effectsOptions.Paste.parent = self;
	self.effectsOptions.Paste.Click = function(self, args)
		local x = self.parent.effectsIndex.x;
		local y = self.parent.effectsIndex.y;
		local t = { };

		t.icon = self.parent.effectsClipboard.icon;
		t.debuff = self.parent.effectsClipboard.debuff;
		t.name = self.parent.effectsClipboard.name;

		if x == nil then
			if self.parent.slotsButton[self.parent.select].missing then
				table.insert(self.parent.slots[self.parent.select], { [1] = t });
			else
				table.insert(self.parent.slots[self.parent.select], t);
			end
		else
			if y == nil then
				table.insert(self.parent.slots[self.parent.select], x, t);
			else
				table.insert(self.parent.slots[self.parent.select][x], y, t);
			end
		end
		self.parent:Changed();
	end;
	items:Add(self.effectsOptions.Paste);
	self.effectsOptions.Delete = Turbine.UI.MenuItem("Delete");
	self.effectsOptions.Delete.parent = self;
	self.effectsOptions.Delete.Click = function(self, args)
		local x = self.parent.effectsIndex.x;
		local y = self.parent.effectsIndex.y;

		if y == nil then
			table.remove(self.parent.slots[self.parent.select], x);
		else
			table.remove(self.parent.slots[self.parent.select][x], y);
			if #self.parent.slots[self.parent.select][x] == 0 then
				table.remove(self.parent.slots[self.parent.select], x);
			end
		end
		self.parent:Changed();
	end;
	items:Add(self.effectsOptions.Delete);
	self.effectsOptions.Insert = Turbine.UI.MenuItem("Insert");
	self.effectsOptions.Insert.parent = self;
	self.effectsOptions.Insert.Click = function(self, args)
		local x = self.parent.effectsIndex.x;
		local y = self.parent.effectsIndex.y;

		if x == nil then
			if self.parent.slotsButton[self.parent.select].missing then
				table.insert(self.parent.slots[self.parent.select], { [1] = { debuff = false } });
			else
				table.insert(self.parent.slots[self.parent.select], { debuff = false });
			end
		else
			if y == nil then
				table.insert(self.parent.slots[self.parent.select], x, { debuff = false });
			else
				table.insert(self.parent.slots[self.parent.select][x], y, { debuff = false });
			end
		end
		self.parent:Changed();
	end;
	items:Add(self.effectsOptions.Insert);
	self.effectsOptions.Debuff = Turbine.UI.MenuItem("Debuff");
	self.effectsOptions.Debuff.parent = self;
	self.effectsOptions.Debuff.Click = function(self, args)
		local x = self.parent.effectsIndex.x;
		local y = self.parent.effectsIndex.y;

		if y == nil then
			self.parent.slots[self.parent.select][x].debuff = not self.parent.slots[self.parent.select][x].debuff;
		else
			self.parent.slots[self.parent.select][x][y].debuff = not self.parent.slots[self.parent.select][x][y].debuff;
		end
		self.parent:Changed();
	end;
	items:Add(self.effectsOptions.Debuff);
	self.effectsOptions.Save = Turbine.UI.MenuItem("Save");
	self.effectsOptions.Save.parent = self;
	self.effectsOptions.Save.Click = function(self, args)
		local x = self.parent.effectsIndex.x;
		local y = self.parent.effectsIndex.y;
		local t;

		if y == nil then
			t = self.parent.slots[self.parent.select][x];
		else
			t = self.parent.slots[self.parent.select][x][y];
		end
		
		for _, v in pairs(self.parent.logSaved) do
			if v.name == t.name and v.debuff == t.debuff and v.icon == t.icon then
				return;
			end
		end
		table.insert(self.parent.logSaved, { debuff = t.debuff, icon = t.icon, name = t.name });
		self.parent:SaveLogSaved();
	end;
	items:Add(self.effectsOptions.Save);
	self.effectsOptions.NewGroup = Turbine.UI.MenuItem("New Group");
	self.effectsOptions.NewGroup.parent = self;
	self.effectsOptions.NewGroup.Click = function(self, args)
		local x = self.parent.effectsIndex.x;
		local y = self.parent.effectsIndex.y;
		local n = #self.parent.slots[self.parent.select][x];

		table.insert(self.parent.slots[self.parent.select], x + 1, { });
		for i = y, n, 1 do
			table.insert(self.parent.slots[self.parent.select][x + 1], self.parent.slots[self.parent.select][x][i]);
			self.parent.slots[self.parent.select][x][i] = nil;
		end
		self.parent:Changed();
	end;
	items:Add(self.effectsOptions.NewGroup);

	self:SetBackColor(Turbine.UI.Color(1.0, 0.0, 0.0, 0.3));

	self:LoadSetupSaved();
	self.setupName = Turbine.UI.Lotro.TextBox();
	self.setupName:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
	self.setupName:SetWantsKeyEvents(true);
	self.setupName.FocusLost = function(self, args)
		for i = -5, 5, 1 do
			if i ~= 0 then
				self.parent.slotsButton[i]:Colorize();
			end
		end
	end;
	self.setupName.KeyDown = function(self, args)
		if self:HasFocus() and args.Action == 162 then
			self.parent:Focus();
		end
	end;
	self.setupName:SetMultiline(false);
	self.setupName:SetFont(Turbine.UI.Lotro.Font.Verdana16);
	self.setupName:SetParent(self);
	self.setupName.parent = self;
	self.setupMenu = Turbine.UI.ContextMenu();
	self.setupDropDown = Turbine.UI.Button();
	self.setupDropDown:SetBackground("moebius92/PartyFrame/dropdown.tga");
	self.setupDropDown:SetSize(window.size - 2, window.size - 2);
	self.setupDropDown:SetParent(self);
	self.setupDropDown.parent = self;
	self.setupDropDown.MouseClick = function(self, args)
		local items = self.parent.setupMenu:GetItems();
		local f = function(self, args) self.parent.setupName:SetText(self:GetText()); self.parent.setupName:FocusLost(); end;
		local item;
		local t = { };

		items:Clear();
		for k in pairs(self.parent.setupSaved) do
			table.insert(t, k);
		end
		table.sort(t);
		if #t > 0 then
			for _, v in ipairs(t) do
				item = Turbine.UI.MenuItem(v);
				item.parent = self.parent;
				item.Click = f;
				items:Add(item);
			end
			self.parent.setupMenu:ShowMenuAt(self.parent.setupName:PointToScreen(0, self.parent.setupName:GetHeight()));
		end
	end;
	self.setupSave = Turbine.UI.Lotro.Button();
	self.setupSave:SetText("Save");
	self.setupSave:SetSize(self.width, window.size - 2);
	self.setupSave:SetParent(self);
	self.setupSave.parent = self;
	self.setupSave.MouseClick = function(self, args)
		local name = self.parent.setupName:GetText();

		if self.parent.select ~= 0 then
			if #self.parent.slots[self.parent.select] > 0 then
				if self.parent.setupSaved[name] == nil then
					self.parent.setupSaved[name] = { };
				end
				self.parent.setupSaved[name][self.parent.select] = { };
				self.parent.setupSaved[name][self.parent.select].effects = tabledeepcopy(self.parent.slots[self.parent.select]);
				self.parent.setupSaved[name][self.parent.select].missing = self.parent.slotsButton[self.parent.select].missing;
				self.parent.setupSaved[name][self.parent.select].duration = self.parent.slotsButton[self.parent.select].duration;
			end
			self.parent.slotsButton[self.parent.select]:Colorize();
			self.parent:SaveSetupSaved();
		end
	end;
	self.setupLoad = Turbine.UI.Lotro.Button();
	self.setupLoad:SetText("Load");
	self.setupLoad:SetSize(self.width, window.size - 2);
	self.setupLoad:SetParent(self);
	self.setupLoad.parent = self;
	self.setupLoad.MouseClick = function(self, args)
		self.parent:LoadSetup(self.parent.setupName:GetText());
	end;
	self.setupDelete = Turbine.UI.Lotro.Button();
	self.setupDelete:SetText("Delete");
	self.setupDelete:SetSize(self.width, window.size - 2);
	self.setupDelete:SetParent(self);
	self.setupDelete.parent = self;
	self.setupDelete.MouseClick = function(self, args)
		local name = self.parent.setupName:GetText();

		if self.parent.select ~= 0 then
			if self.parent.setupSaved[name] ~= nil then
				self.parent.setupSaved[name][self.parent.select] = nil;
				if next(self.parent.setupSaved[name]) == nil then
					self.parent.setupSaved[name] = nil;
				end
				self.parent:SaveSetupSaved();
			end
			self.parent.slotsButton[self.parent.select]:Colorize();
		end
	end;

	self.slots = self:LoadSlots();
	self.slotsButton = { };
	for i = -5, 5, 1 do
		if i ~= 0 then
			self.slotsButton[i] = PartyFrameOptionsButton(self, i, self.slots[i].missing);
			self.slotsButton[i].duration = self.slots[i].duration;
		end
		if self.slots[i].effects == nil then
			self:Default(i);
		else
			self.slots[i] = self.slots[i].effects;
		end
	end
	self.slotsType = PartyFrameOptionsDropDown(self, 1, function(self, index) self.parent.slotsButton[self.parent.select]:SetMissing(index == 2); end, "Existent", "Missing");
	self.slotsType:SetSize(self.width, window.size - 2);
	self.slotsDuration = Turbine.UI.Lotro.TextBox();
	self.slotsDuration:SetSize(self.width / 3, window.size - 2);
	self.slotsDuration:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleRight);
	self.slotsDuration:SetWantsKeyEvents(true);
	self.slotsDuration.FocusGained = function(self, args) self:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft); end;
	self.slotsDuration.FocusLost = function(self, args) self:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleRight); end;
	self.slotsDuration.KeyDown = function(self, args)
		if self:HasFocus() and args.Action == 162 then
			local n = tonumber(self:GetText());

			self.parent:Focus();
			if n ~= nil and n >= 0 and math.floor(n) == n then
				self.parent.slotsButton[self.parent.select].duration = n;
				self.parent:Changed();
			end
			self:SetText(self.parent.slotsButton[self.parent.select].duration);
		end
	end;
	self.slotsDuration:SetMultiline(false);
	self.slotsDuration:SetFont(Turbine.UI.Lotro.Font.Verdana16);
	self.slotsDuration:SetParent(self);
	self.slotsDuration.parent = self;
	self.slotsDurationLabel = Turbine.UI.Label();
	self.slotsDurationLabel:SetSize(self.width * 2 / 3, window.size - 2);
	self.slotsDurationLabel:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
	self.slotsDurationLabel:SetFont(Turbine.UI.Lotro.Font.Verdana16);
	self.slotsDurationLabel:SetText("seconds");
	self.slotsDurationLabel:SetParent(self);
	self:UpdateSlots();

	self:LoadLogSaved();
	for k, v in pairs(self.logSaved) do
		self.slots[0][k] = v;
	end
	self.logButton = Turbine.UI.Lotro.Button();
	self.logButton.parent = self;
	self.logButton:SetParent(self);
	self.logButton:SetText("Log");
	self.logButton:SetSize(self.width, window.size - 2);
	self.logButton.MouseClick = function(self, args) self.parent:Select(0); end;
	self.logDebuffs = PartyFrameOptionsDropDown(self, 3, nil, "No Debuffs", "No Icon Debuffs", "Unknown Debuffs", "All Debuffs");
	self.logDebuffs:SetSize(self.width * 1.5, window.size - 2);
	self.logBuffs = PartyFrameOptionsDropDown(self, 1, nil, "No Buffs", "No Icon Buffs", "Unknown Buffs", "All Buffs");
	self.logBuffs:SetSize(self.width * 1.5, window.size - 2);

	self.logClear = Turbine.UI.Lotro.Button();
	self.logClear.parent = self;
	self.logClear:SetParent(self);
	self.logClear:SetText("Clear");
	self.logClear:SetSize(self.width, window.size - 2);
	self.logClear.MouseClick = function(self, args)
		self.parent.slots[self.parent.select] = { };
		if self.parent.select == 0 then
			self.parent.logSaved = { };
		end
		self.parent:Changed();
	end;
	self.logDefault = Turbine.UI.Lotro.Button();
	self.logDefault.parent = self;
	self.logDefault:SetParent(self);
	self.logDefault:SetText("Default");
	self.logDefault:SetSize(self.width, window.size - 2);
	self.logDefault.MouseClick = function(self, args)
		if self.parent.select == 0 then
			self.parent.slots[0] = { };
			for k, v in pairs(self.parent.logSaved) do
				self.parent.slots[0][k] = v;
			end
		else
			self.parent:Default(self.parent.select);
		end
		self.parent:Changed();
	end;

	self.dragWindow = Turbine.UI.Window();
	self.dragWindow:SetSize(32, 32);
	self.dragWindow:SetBackColor(Turbine.UI.Color(1.0, 0.0, 0.0, 0.0));
	self.dragWindow:SetOpacity(0.7);
end

combatStates = { };
combatStates["Dazed"] = 0x410D8360;
combatStates["Feared"] = 0x410D7BA0;
combatStates["Knocked Down"] = 0x4100823F;
combatStates["Knocked Out"] = 0x4100823F;
combatStates["Recovering"] = 0x410E7686;
combatStates["Rooted"] = 0x4100823D;
combatStates["Stunned"] = 0x4100823F;
combatStates["Vulnerable"] = 0x41008214;

debuffSlot = { };
debuffSlot[-4] = "<fear-curable>";
debuffSlot[-3] = "<wound-curable>";
debuffSlot[-2] = "<disease-curable>";
debuffSlot[-1] = "<poison-curable>";

buffSlot = { }
buffSlot[Turbine.Gameplay.Class.Burglar] = { };
buffSlot[Turbine.Gameplay.Class.Burglar][1] = { };
buffSlot[Turbine.Gameplay.Class.Burglar][1]["Signal: Conviction"] = 0x410DAF06;
buffSlot[Turbine.Gameplay.Class.Burglar][1]["Signal: Guile"] = 0x410DAF0A;
buffSlot[Turbine.Gameplay.Class.Burglar][1]["Signal: Strength"] = 0x410DAF02;
buffSlot[Turbine.Gameplay.Class.Burglar][1]["Signal: Tactics"] = 0x410DAF0E;

buffSlot[Turbine.Gameplay.Class.Captain] = { };
buffSlot[Turbine.Gameplay.Class.Captain][1] = { };
buffSlot[Turbine.Gameplay.Class.Captain][1]["Blade-brother"] = 0x411167F0;
buffSlot[Turbine.Gameplay.Class.Captain][1]["Shield-brother"] = 0x41104588;
buffSlot[Turbine.Gameplay.Class.Captain][1]["Song-brother"] = 0x411167EF;
buffSlot[Turbine.Gameplay.Class.Captain][1]["Watchful Shield-brother"] = 0x41104588;
buffSlot[Turbine.Gameplay.Class.Captain][2] = { };
buffSlot[Turbine.Gameplay.Class.Captain][2]["To Arms (Blade-brother)"] = 0x4107E9F2;
buffSlot[Turbine.Gameplay.Class.Captain][2]["To Arms (Shield-brother)"] = 0x411167F5;
buffSlot[Turbine.Gameplay.Class.Captain][2]["To Arms (Song-brother)"] = 0x411167F6;
buffSlot[Turbine.Gameplay.Class.Captain][2]["To Arms (Fellowship Blade-brother)"] = 0x4107E9F2;
buffSlot[Turbine.Gameplay.Class.Captain][2]["To Arms (Fellowship Shield-brother)"] = 0x411167F5;
buffSlot[Turbine.Gameplay.Class.Captain][2]["To Arms (Fellowship Song-brother)"] = 0x411167F6;
buffSlot[Turbine.Gameplay.Class.Captain][3] = { };
buffSlot[Turbine.Gameplay.Class.Captain][3]["War-cry"] = 0x4100560F;
buffSlot[Turbine.Gameplay.Class.Captain][4] = { };
buffSlot[Turbine.Gameplay.Class.Captain][4]["Make Haste"] = 0x4100371F;
buffSlot[Turbine.Gameplay.Class.Captain][4]["Rallying Cry"] = 0x41005609;
buffSlot[Turbine.Gameplay.Class.Captain][4]["Valiant Strike"] = 0x410E68E8;
buffSlot[Turbine.Gameplay.Class.Captain][4]["Words of Courage"] = 0x4100561B;

buffSlot[Turbine.Gameplay.Class.Champion] = { };

buffSlot[Turbine.Gameplay.Class.Guardian] = { };
buffSlot[Turbine.Gameplay.Class.Guardian][1] = { };
buffSlot[Turbine.Gameplay.Class.Guardian][1]["Protection"] = 0x41005678;
buffSlot[Turbine.Gameplay.Class.Guardian][1]["Protection by the Sword"] = 0x4111735D;
buffSlot[Turbine.Gameplay.Class.Guardian][1]["Shield Wall"] = 0x410DAEF3;

buffSlot[Turbine.Gameplay.Class.Hunter] = { };
buffSlot[Turbine.Gameplay.Class.Hunter][1] = { };
buffSlot[Turbine.Gameplay.Class.Hunter][1]["Find the Path"] = 0x41008858;
buffSlot[Turbine.Gameplay.Class.Hunter][2] = { };
buffSlot[Turbine.Gameplay.Class.Hunter][2]["Cleanse Poison Resistance"] = 0x41008859;

buffSlot[Turbine.Gameplay.Class.LoreMaster] = { };
buffSlot[Turbine.Gameplay.Class.LoreMaster][1] = { };
buffSlot[Turbine.Gameplay.Class.LoreMaster][1]["Charged Air"] = 0x4110458C;
buffSlot[Turbine.Gameplay.Class.LoreMaster][2] = { };
buffSlot[Turbine.Gameplay.Class.LoreMaster][2]["Sign of Power: Righteousness"] = 0x4100881A;
buffSlot[Turbine.Gameplay.Class.LoreMaster][3] = { };
buffSlot[Turbine.Gameplay.Class.LoreMaster][3]["Sign of Power: Vigilance"] = 0x4100881B;

buffSlot[Turbine.Gameplay.Class.Minstrel] = { };
buffSlot[Turbine.Gameplay.Class.Minstrel][1] = { };
buffSlot[Turbine.Gameplay.Class.Minstrel][1]["Soliloquy of Spirit"] = 0x410E3A1F;
buffSlot[Turbine.Gameplay.Class.Minstrel][2] = { };
buffSlot[Turbine.Gameplay.Class.Minstrel][2]["Enlivening Chord"] = 0x41004E8F;
buffSlot[Turbine.Gameplay.Class.Minstrel][2]["Inspire Fellows Buff"] = 0x410087C7;
buffSlot[Turbine.Gameplay.Class.Minstrel][2]["The Melody of Battle"] = 0x410D73F5;
buffSlot[Turbine.Gameplay.Class.Minstrel][2]["Valar's Grace"] = 0x4100A4E2;
buffSlot[Turbine.Gameplay.Class.Minstrel][2]["Words of Courage: Fear Resistance"] = 0x410087D5;
buffSlot[Turbine.Gameplay.Class.Minstrel][3] = { };
buffSlot[Turbine.Gameplay.Class.Minstrel][3]["Fellowship's Heart"] = 0x4100A4E2;
buffSlot[Turbine.Gameplay.Class.Minstrel][4] = { };
buffSlot[Turbine.Gameplay.Class.Minstrel][4]["Tale of Battle"] = 0x410D715F;
buffSlot[Turbine.Gameplay.Class.Minstrel][4]["Tale of Frost and Flame"] = 0x410E3A26;
buffSlot[Turbine.Gameplay.Class.Minstrel][4]["Tale of Heroism"] = 0x410087D2;
buffSlot[Turbine.Gameplay.Class.Minstrel][4]["Tale of Warding"] = 0x410D7160;
buffSlot[Turbine.Gameplay.Class.Minstrel][4]["Tale of Warding and Hope"] = 0x410D7160;
buffSlot[Turbine.Gameplay.Class.Minstrel][4]["Improved Tale of Battle"] = 0x410D715F;
buffSlot[Turbine.Gameplay.Class.Minstrel][4]["Improved Tale of Warding"] = 0x410D7160;
buffSlot[Turbine.Gameplay.Class.Minstrel][5] = { };
buffSlot[Turbine.Gameplay.Class.Minstrel][5]["Anthem of Composure"] = 0x410087C6;
buffSlot[Turbine.Gameplay.Class.Minstrel][5]["Anthem of the Free People"] = 0x41005618;
buffSlot[Turbine.Gameplay.Class.Minstrel][5]["Anthem of Prowess"] = 0x410087C5;
buffSlot[Turbine.Gameplay.Class.Minstrel][5]["Anthem of War"] = 0x410056AE;

buffSlot[Turbine.Gameplay.Class.RuneKeeper] = { };
buffSlot[Turbine.Gameplay.Class.RuneKeeper][1] = { };
buffSlot[Turbine.Gameplay.Class.RuneKeeper][1]["Improved Rune-sign of Flame"] = 0x4110458E;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][1]["Improved Rune-sign of Storm"] = 0x41104590;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][1]["Improved Rune-sign of Winter"] = 0x4110458F;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][1]["Prelude to Hope"] = 0x410E44D7;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][1]["Rune-sign of Flame"] = 0x4110458E;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][1]["Rune-sign of Storm"] = 0x41104590;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][1]["Rune-sign of Winter"] = 0x4110458F;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][2] = { };
buffSlot[Turbine.Gameplay.Class.RuneKeeper][2]["Writ of Health - Tier 1"] = 0x410EA25B;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][2]["Writ of Health - Tier 2"] = 0x410EA25C;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][2]["Writ of Health - Tier 3"] = 0x410EA25A;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][3] = { };
buffSlot[Turbine.Gameplay.Class.RuneKeeper][3]["Mending Verse"] = 0x410E44D4;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][4] = { };
buffSlot[Turbine.Gameplay.Class.RuneKeeper][4]["Essay of Exaltation"] = 0x410F3580;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][4]["Glorious Foreshadowing"] = 0x410F23E9;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][4]["Rousing Words"] = 0x410E44CE;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][4]["That Which Does Not Kill Us"] = 0x410F23E6;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][4]["Wondrous Foreshadowing"] = 0x410F23EA;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][5] = { };
buffSlot[Turbine.Gameplay.Class.RuneKeeper][5]["Do not fall to Fire"] = 0x410EA22F;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][5]["Do not fall to Frost"] = 0x410EA234;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][5]["Do not fall to Lightning"] = 0x410EA235;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][5]["Do Not Fall This Day"] = 0x410E44DC;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][5]["Shall Not Fall This Day"] = 0x410E44DC;
buffSlot[Turbine.Gameplay.Class.RuneKeeper][5]["Their Weapons Shall Not Harm Us"] = 0;

buffSlot[Turbine.Gameplay.Class.Warden] = { };
buffSlot[Turbine.Gameplay.Class.Warden][1] = { };
buffSlot[Turbine.Gameplay.Class.Warden][1]["Conviction"] = 0x410EA2A7;
buffSlot[Turbine.Gameplay.Class.Warden][2] = { };
buffSlot[Turbine.Gameplay.Class.Warden][2]["Conviction - Physical Mastery"] = 0x410E8292;
buffSlot[Turbine.Gameplay.Class.Warden][2]["Conviction - Threat Reduction"] = 0x410E87A3;

captainSlot = { };
captainSlot[1] = { };
captainSlot[1][1] = { name = "Focus", icon = 0x410087BB };
captainSlot[1][2] = { name = "On Guard", icon = 0x410087B5 };
captainSlot[1][3] = { name = "Relentless Attack", icon = 0x410087B7 };
captainSlot[2] = { };
captainSlot[2][1] = { name = "Motivated", icon = 0x410087B4 };
captainSlot[3] = { };
captainSlot[3][1] = { name = "Strength of Will (Shield-brother)", icon = 0x410D3FFA };
captainSlot[3][2] = { name = "Strength of Will (Blade-brother)", icon = 0x411167F4 };
captainSlot[3][3] = { name = "Strength of Will (Song-brother)", icon = 0x411167F3 };
captainSlot[3][4] = { name = "Strength of Will (Fellowship Blade-brother)", icon = 0x411167F4 };
captainSlot[3][5] = { name = "Strength of Will (Fellowship Shield-brother)", icon = 0x410D3FFA };
captainSlot[3][6] = { name = "Strength of Will (Fellowship Song-brother)", icon = 0x411167F3 };
captainSlot[4] = { };
captainSlot[4][1] = { name = "Inspired (Blade-brother)", icon = 0x411167F2 };
captainSlot[4][2] = { name = "Inspired (Shield-brother)", icon = 0x4107E9F1 };
captainSlot[4][3] = { name = "Inspired (Song-brother)", icon = 0x411167F1 };
captainSlot[4][4] = { name = "Inspired (Fellowship Blade-brother)", icon = 0x411167F2 };
captainSlot[4][5] = { name = "Inspired (Fellowship Shield-brother)", icon = 0x4107E9F1 };
captainSlot[4][6] = { name = "Inspired (Fellowship Song-brother)", icon = 0x411167F1 };
captainSlot[5] = { };
captainSlot[5][1] = { name = "Light of Elendil", icon = 0x4109B53E };

function PartyFrameOptions:Default(i)
	self.slots[i] = { };
	if i ~= 0 then
		self.slotsButton[i].missing = false;
		if i < 0 then
			self.slotsButton[i].duration = 10;
		else
			self.slotsButton[i].duration = 0;
		end
	end
	if i == -5 then
		for k, v in pairs(combatStates) do
			local d = (k ~= "Recovering");

			if v ~= 0 then
				table.insert(self.slots[i], { icon = v, debuff = d, name = k });
			else
				table.insert(self.slots[i], { debuff = d, name = k });
			end
		end
	elseif i < 0 then
		table.insert(self.slots[i], { debuff = true, name = debuffSlot[i] });
	elseif i == 5 and self.window.class == Turbine.Gameplay.Class.Captain then
		self.slotsButton[i].missing = true;
		for k, v in ipairs(captainSlot) do
			self.slots[i][k] = { };
			for _, v1 in ipairs(v) do
				table.insert(self.slots[i][k], { icon = v1.icon, debuff = false, name = v1.name });
			end
		end
	elseif buffSlot[self.window.class][i] ~= nil then
		for k, v in pairs(buffSlot[self.window.class][i]) do
			if v ~= 0 then
				table.insert(self.slots[i], { icon = v, debuff = false, name = k });
			else
				table.insert(self.slots[i], { debuff = false, name = k });
			end
		end
	end

	if i ~= 0 then
		self.slotsButton[i]:Colorize();
	end
	if i == 0 or not self.slotsButton[i].missing then
		table.sort(self.slots[i], function(a, b) return (not a.name and "" or a.name) < (not b.name and "" or b.name); end);
	end
end

function PartyFrameOptions:Changed()
	self:SizeChanged();
	self:SaveLogSaved();
	self:SaveSlots();
	self:UpdateSlots();
end

function PartyFrameOptions:UpdateSlots()
	local name;

	effectGroup = { };
	effectSlot = { };

	for i = -5, 5, 1 do
		if i ~= 0 then
			if self.slotsButton[i].missing then
				effectGroup[i] = { };
				for k, v in ipairs(self.slots[i]) do
					effectGroup[i][k] = { };
					effectGroup[i][k].icon = v[1].icon;
					effectGroup[i][k].effects = { };

					for _, v1 in pairs(v) do
						if v1.debuff then
							name = "D";
						else
							name = "B";
						end

						if v1.icon ~= nil then
							name = name .. "/" .. v1.icon;
						end
						if v1.name ~= nil then
							name = name .. "/" .. v1.name;
						end

						if string.len(name) > 1 then
							effectSlot[name] = i;
							effectGroup[i][k].effects[name] = true;
						end
					end
				end
			else
				for _, v in pairs(self.slots[i]) do
					if v.debuff then
						name = "D";
					else
						name = "B";
					end

					if v.icon ~= nil then
						name = name .. "/" .. v.icon;
					end
					if v.name ~= nil then
						name = name .. "/" .. v.name;
					end

					if string.len(name) > 1 then
						effectSlot[name] = i;
					end
				end
			end
		end
	end

	for k, v in pairs(self.window.actors) do
		v:SetupEffects();
	end
end

function PartyFrameOptions:SaveSlots()
	local t = { };

	for k, v in pairs(self.slots) do
		if k ~= 0 then
			t[k] = { };
			t[k].missing = self.slotsButton[k].missing;
			t[k].duration = self.slotsButton[k].duration;
			t[k].effects = v;
		end
	end
	t.version = self.window.version;
	Turbine.PluginData.Save(Turbine.DataScope.Character, "PartyFrame_Slots", serialize(t));
end

function PartyFrameOptions:LoadSlots()
	local t = Turbine.PluginData.Load(Turbine.DataScope.Character, "PartyFrame_Slots");
	if type(t) == "string" then
		t = loadstring("return " .. t .. ";")();
	end

	if t == nil or t.version ~= self.window.version then
		t = { };
	end

	for i = -5, 5, 1 do
		if t[i] == nil then
			t[i] = { };
		end
	end

	t.version = nil;

	return t;
end

function PartyFrameOptions:SaveLogSaved()
	self.logSaved.version = self.window.version;
	Turbine.PluginData.Save(Turbine.DataScope.Account, "PartyFrame_LogSaved", serialize(self.logSaved));
	self.logSaved.version = nil;
end

function PartyFrameOptions:LoadLogSaved()
	self.logSaved = Turbine.PluginData.Load(Turbine.DataScope.Account, "PartyFrame_LogSaved");
	if type(self.logSaved) == "string" then
		self.logSaved = loadstring("return " .. self.logSaved .. ";")();
	end
	if self.logSaved == nil or self.logSaved.version ~= self.window.version then
		self.logSaved = { };
	end
	self.logSaved.version = nil;
end

function PartyFrameOptions:SaveSetupSaved()
	self.setupSaved.version = self.window.version;
	Turbine.PluginData.Save(Turbine.DataScope.Account, "PartyFrame_SetupSaved", serialize(self.setupSaved));
	self.setupSaved.version = nil;
end

function PartyFrameOptions:LoadSetupSaved()
	self.setupSaved = Turbine.PluginData.Load(Turbine.DataScope.Account, "PartyFrame_SetupSaved");
	if type(self.setupSaved) == "string" then
		self.setupSaved = loadstring("return " .. self.setupSaved .. ";")();
	end
	if self.setupSaved == nil or self.setupSaved.version ~= self.setupSaved.version then
		self.setupSaved = { };
	end
	self.setupSaved.version = nil;
end

function PartyFrameOptions:LoadSetup(name)
	if self.setupSaved[name] ~= nil then
		self.setupName:SetText(name);
		for i = -5, 5, 1 do
			if self.setupSaved[name][i] ~= nil then
				self.slots[i] = tabledeepcopy(self.setupSaved[name][i].effects);
				self.slotsButton[i].missing = self.setupSaved[name][i].missing;
				self.slotsButton[i].duration = self.setupSaved[name][i].duration;
			end
		end
		self:Select(self.select);
		self:Changed();
		return true;
	end
	return false;
end

function PartyFrameOptions:GetIndices(i)
	local x = nil;
	local y = nil;

	if self.select == 0 then
		x = i;
	elseif self.slotsButton[self.select].missing then
		for k, v in ipairs(self.slots[self.select]) do
			if i <= #v then
				x = k;
				y = i;
				break;
			end
			i = i - #v;
		end
	else
		if self.slots[self.select][i] ~= nil then
			x = i;
		end
	end

	return x, y;
end

function PartyFrameOptions:Select(i)
	self.select = i;

	for _, v in pairs(self.slotsButton) do
		v:Colorize();
	end

	if self.select ~= 0 then
		if self.slotsButton[self.select].missing then
			self.slotsType:SetIndex(2);
		else
			self.slotsType:SetIndex(1);
		end
		self.slotsDuration:SetText(self.slotsButton[self.select].duration);
	end

	self:SizeChanged();
end

function PartyFrameOptions:SizeChanged(args)
	local w, h = self:GetSize();
	local x;
	local y = 0;

	y = y + 10;
	x = w - (self.window.size * 10 + self.width);
	if x > 0 then
		x = x / 2;
		for i = -5, 5, 1 do
			if i == 0 then
				self.logButton:SetPosition(x, y);
				x = x + self.width;
			else
				self.slotsButton[i]:SetPosition(x, y);
				x = x + self.window.size;
			end
		end
		y = y + self.window.size;
	else
		x = (w - (self.window.size * 5)) / 2;
		for i = -5, -1, 1 do
			self.slotsButton[i]:SetPosition(x, y);
			x = x + self.window.size;
		end
		y = y + self.window.size;

		x = (w - self.width) / 2;
		self.logButton:SetPosition(x, y);
		y = y + self.window.size;

		x = (w - (self.window.size * 5)) / 2;
		for i = 1, 5, 1 do
			self.slotsButton[i]:SetPosition(x, y);
			x = x + self.window.size;
		end
		y = y + self.window.size;
	end
	y = y + 10;

	if self.select == 0 then
		self.logBuffs:SetVisible(true);
		self.logDebuffs:SetVisible(true);
		self.slotsType:SetVisible(false);
		self.slotsDuration:SetVisible(false);
		self.slotsDurationLabel:SetVisible(false);
		x = w - (self.width * 3 + 10);
		if x > 0 then
			x = x / 2;
			self.logDebuffs:SetPosition(x, y);
			x = x + self.width * 1.5 + 5
			self.logBuffs:SetPosition(x, y);
			y = y + self.window.size;
		else
			x = (w - self.width * 1.5) / 2;
			self.logDebuffs:SetPosition(x, y);
			y = y + self.window.size;
			self.logBuffs:SetPosition(x, y);
			y = y + self.window.size;
		end
	else
		self.logBuffs:SetVisible(false);
		self.logDebuffs:SetVisible(false);
		self.slotsType:SetVisible(true);
		self.slotsDuration:SetVisible(true);
		self.slotsDurationLabel:SetVisible(true);
		x = w - (self.width * 2 + 5);
		if x > 0 then
			x = x / 2;
			self.slotsType:SetPosition(x, y);
			x = x + self.width + 5;
			self.slotsDuration:SetPosition(x, y);
			x = x + self.width / 3;
			self.slotsDurationLabel:SetPosition(x, y);
			y = y + self.window.size;
		else
			x = (w - self.width) / 2;
			self.slotsType:SetPosition(x, y);
			y = y + self.window.size;
			x = (w - self.width) / 2;
			self.slotsDuration:SetPosition(x, y);
			x = x + self.width / 3;
			self.slotsDurationLabel:SetPosition(x, y);
			y = y + self.window.size;
		end
	end
	y = y + 10;

	for k, v in ipairs(self.effects) do
		v:SetVisible(false);
	end
	for k, v in ipairs(self.groups) do
		v:SetVisible(false);
	end
	if self.select == 0 then
		for k, v in ipairs(self.slots[0]) do
			if self.effects[k] == nil then
				self.effects[k] = PartyFrameOptionsEffect(self, k, v.icon, v.debuff, v.name, true);
			else
				self.effects[k]:SetIcon(v.icon);
				self.effects[k]:SetDebuff(v.debuff);
				self.effects[k]:SetName(v.name, true);
			end
			self.effects[k]:SetSize(w - (6 + self.window.size), self.window.size + 2);
			self.effects[k]:SetPosition(4 + self.window.size, y);
			self.effects[k]:SetVisible(true);
			y = y + self.window.size + 2;
		end
	else
		local i = 1;

		if self.slotsButton[self.select].missing then
			for k, v in ipairs(self.slots[self.select]) do
				if self.groups[k] == nil then
					self.groups[k] = PartyFrameOptionsGroup(self, k);
				end
				self.groups[k]:Setup(#v);
				self.groups[k]:SetPosition(2, y + 2);

				for _, v1 in ipairs(v) do
					if self.effects[i] == nil then
						self.effects[i] = PartyFrameOptionsEffect(self, i, v1.icon, v1.debuff, v1.name, false);
					else
						self.effects[i]:SetIcon(v1.icon);
						self.effects[i]:SetDebuff(v1.debuff);
						self.effects[i]:SetName(v1.name, false);
					end
					self.effects[i]:SetSize(w - (6 + self.window.size), self.window.size + 2);
					self.effects[i]:SetPosition(4 + self.window.size, y);
					self.effects[i]:SetVisible(true);
					i = i + 1;
					y = y + self.window.size + 2;
				end
			end
		else
			for _, v in ipairs(self.slots[self.select]) do
				if self.effects[i] == nil then
					self.effects[i] = PartyFrameOptionsEffect(self, i, v.icon, v.debuff, v.name, false);
				else
					self.effects[i]:SetIcon(v.icon);
					self.effects[i]:SetDebuff(v.debuff);
					self.effects[i]:SetName(v.name, false);
				end
				self.effects[i]:SetSize(w - (6 + self.window.size), self.window.size + 2);
				self.effects[i]:SetPosition(4 + self.window.size, y);
				self.effects[i]:SetVisible(true);
				i = i + 1;
				y = y + self.window.size + 2;
			end
		end

		if self.effects[i] == nil then
			self.effects[i] = PartyFrameOptionsEffect(self, i, nil, false, nil, false);
		else
			self.effects[i]:SetIcon(nil);
			self.effects[i]:SetDebuff(false);
			self.effects[i]:SetName(nil, false);
		end
		self.effects[i]:SetSize(w - (6 + self.window.size), self.window.size + 2);
		self.effects[i]:SetPosition(4 + self.window.size, y);
		self.effects[i]:SetVisible(true);
		i = i + 1;
		y = y + self.window.size + 2;
	end
	y = y + 10;

	x = w - (self.width * 2 + 5);
	if x > 0 then
		x = x / 2;
		self.logClear:SetPosition(x, y);
		x = x + self.width + 5
		self.logDefault:SetPosition(x, y);
		y = y + self.window.size;
	else
		x = (w - self.width) / 2;
		self.logClear:SetPosition(x, y);
		y = y + self.window.size;
		self.logDefault:SetPosition(x, y);
		y = y + self.window.size;
	end
	y = y + 10;

	local l = math.min(w - 4, self.width * 3 + 10);
	x = (w - l) / 2;
	self.setupName:SetSize(l - self.window.size, self.window.size - 2);
	self.setupName:SetPosition(x, y);
	x = x + (l - self.window.size);
	self.setupDropDown:SetPosition(x, y);
	y = y + self.window.size;
	x = w - (self.width * 3 + 10);
	if x > 0 then
		x = x / 2;
		self.setupSave:SetPosition(x, y);
		x = x + self.width + 5;
		self.setupLoad:SetPosition(x, y);
		x = x + self.width + 5;
		self.setupDelete:SetPosition(x, y);
		y = y + self.window.size;
	else
		x = (w - self.width) / 2;
		self.setupSave:SetPosition(x, y);
		y = y + self.window.size;
		self.setupLoad:SetPosition(x, y);
		y = y + self.window.size;
		self.setupDelete:SetPosition(x, y);
		y = y + self.window.size;
	end
	y = y + 10;

	self:SetHeight(y);
end

PartyFrameCommand = class(Turbine.ShellCommand);

function PartyFrameCommand:Constructor(plugin)
	Turbine.ShellCommand.Constructor(self);

	self.plugin = plugin;
	self.window = plugin.window;
end

function PartyFrameCommand:Execute(command, arguments)
	local args = snarfargs(arguments);

	if args[1] == "clear" then
		for k, v in pairs(self.window.players) do
			v.rolls = { };
			v.rollOrder = nil;
			v:UpdateName();
		end
	elseif args[1] == "load" then
		if self.window.options:LoadSetup(args[2]) then
			Turbine.Shell.WriteLine("PartyFrame: " .. args[2] .. " loaded.");
		else
			Turbine.Shell.WriteLine("PartyFrame: " .. args[2] .. " not found.");
		end
	elseif args[1] == "options" then
		Turbine.PluginManager.ShowOptions(self.plugin);
	else
		Turbine.Shell.WriteLine("Usage: /pf <clear|load <name>|options>");
	end
end

function PartyFrameCommand:GetShortHelp()
	return "PartyFrame";
end

plugin.window = PartyFrameWindow(22, 100, plugin:GetVersion());
plugin.options = plugin.window.options;
plugin.command = PartyFrameCommand(plugin);
Turbine.Shell.AddCommand("partyframe", plugin.command);
Turbine.Shell.AddCommand("pf", plugin.command);

function plugin:Load(args)
	local state = Turbine.PluginData.Load(Turbine.DataScope.Account, "PartyFrame");
	if type(state) == "string" then
		state = loadstring("return " .. state .. ";")();
	end
	if state ~= nil and (Turbine.Engine.GetLocalTime() - state.time) < 15 then
		self.window.fellowship = state.fellowship;
		self.window.grid = state.grid;
		self.window:PartyMemberChanged();
	end
end

function plugin:Unload(args)
	local state = { };

	state.time = Turbine.Engine.GetLocalTime();
	state.fellowship = self.window.fellowship;
	state.grid = self.window.grid;
	Turbine.PluginData.Save(Turbine.DataScope.Account, "PartyFrame", serialize(state));
end

function plugin:GetOptionsPanel()
	return self.options;
end